# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for Workflow resources."""
import uuid

import structlog
from flask import request, send_file
from flask_accepts import accepts
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import inject
from structlog.stdlib import BoundLogger

from .schema import FileTypes, JobFilesDownloadQueryParametersSchema
from .service import JobFilesDownloadService

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace("Workflows", description="Workflows endpoint")


@api.route("/jobFilesDownload")
class JobFilesDownloadEndpoint(Resource):
    @inject
    def __init__(
        self, job_files_download_service: JobFilesDownloadService, *args, **kwargs
    ) -> None:
        """Initialize the workflow resource.

        All arguments are provided via dependency injection.

        Args:
            job_files_download_service: A JobFilesDownloadService object.
        """
        self._job_files_download_service = job_files_download_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=JobFilesDownloadQueryParametersSchema, api=api)
    def get(self):
        """Download a compressed file archive containing the files needed to execute a submitted job."""  # noqa: B950
        log = LOGGER.new(  # noqa: F841
            request_id=str(uuid.uuid4()),
            resource="JobFilesDownload",
            request_type="GET",
        )
        mimetype = {
            FileTypes.TAR_GZ: "application/gzip",
            FileTypes.ZIP: "application/zip",
        }
        download_name = {
            FileTypes.TAR_GZ: "job_files.tar.gz",
            FileTypes.ZIP: "job_files.zip",
        }
        parsed_query_params = request.parsed_query_params  # noqa: F841
        job_files_download_package = self._job_files_download_service.get(
            job_id=parsed_query_params["job_id"],
            file_type=parsed_query_params["file_type"],
            log=log,
        )
        return send_file(
            path_or_file=job_files_download_package.name,
            as_attachment=True,
            mimetype=mimetype[parsed_query_params["file_type"]],
            download_name=download_name[parsed_query_params["file_type"]],
        )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import datetime
from typing import TYPE_CHECKING

from sqlalchemy import (
    BigInteger,
    Column,
    ForeignKey,
    ForeignKeyConstraint,
    Index,
    Integer,
    Nullable,
    Text,
    UniqueConstraint,
    and_,
    func,
    select,
)
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import Mapped, column_property, mapped_column, relationship

from dioptra.restapi.db.db import (
    bigint,
    datetimetz,
    db,
    intpk,
    json_,
    optionalbigint,
    optionalstr,
    text_,
)

from .constants import resource_lock_types
from .locks import ResourceLock

if TYPE_CHECKING:
    from .groups import Group
    from .tags import Tag
    from .users import User

# -- Tables (no ORM) -------------------------------------------------------------------

resource_types_table = db.Table(
    "resource_types",
    Column("resource_type", Text(), primary_key=True),
)
resource_dependencies_table = db.Table(
    "resource_dependencies",
    Column(
        "parent_resource_id",
        BigInteger().with_variant(Integer, "sqlite"),
        primary_key=True,
    ),
    Column(
        "child_resource_id",
        BigInteger().with_variant(Integer, "sqlite"),
        primary_key=True,
    ),
    Column(
        "parent_resource_type",
        Text(),
        nullable=False,
    ),
    Column(
        "child_resource_type",
        Text(),
        nullable=False,
    ),
    Index(None, "child_resource_id", "parent_resource_id"),
    Index(None, "parent_resource_type", "child_resource_type"),
    ForeignKeyConstraint(
        ["parent_resource_type", "child_resource_type"],
        [
            "resource_dependency_types.parent_resource_type",
            "resource_dependency_types.child_resource_type",
        ],
    ),
    ForeignKeyConstraint(
        ["parent_resource_id", "parent_resource_type"],
        [
            "resources.resource_id",
            "resources.resource_type",
        ],
    ),
    ForeignKeyConstraint(
        ["child_resource_id", "child_resource_type"],
        [
            "resources.resource_id",
            "resources.resource_type",
        ],
    ),
)
resource_dependency_types_table = db.Table(
    "resource_dependency_types",
    Column(
        "parent_resource_type",
        Text(),
        ForeignKey("resource_types.resource_type"),
        primary_key=True,
    ),
    Column(
        "child_resource_type",
        Text(),
        ForeignKey("resource_types.resource_type"),
        primary_key=True,
    ),
)
resource_tags_table = db.Table(
    "resource_tags",
    Column(
        "resource_id",
        BigInteger().with_variant(Integer, "sqlite"),
        ForeignKey("resources.resource_id"),
        primary_key=True,
    ),
    Column(
        "tag_id",
        BigInteger().with_variant(Integer, "sqlite"),
        ForeignKey("tags.tag_id"),
        primary_key=True,
    ),
)
shared_resource_tags_table = db.Table(
    "shared_resource_tags",
    Column(
        "shared_resource_id",
        BigInteger().with_variant(Integer, "sqlite"),
        ForeignKey("shared_resources.shared_resource_id"),
        primary_key=True,
    ),
    Column(
        "tag_id",
        BigInteger().with_variant(Integer, "sqlite"),
        ForeignKey("tags.tag_id"),
        primary_key=True,
    ),
)

# -- ORM Classes -----------------------------------------------------------------------


class DraftResource(db.Model):  # type: ignore[name-defined]
    __tablename__ = "draft_resources"

    # Database fields
    draft_resource_id: Mapped[intpk] = mapped_column(init=False)
    group_id: Mapped[bigint] = mapped_column(
        ForeignKey("groups.group_id"), init=False, nullable=False
    )
    resource_type: Mapped[text_] = mapped_column(
        ForeignKey("resource_types.resource_type"), nullable=False
    )
    user_id: Mapped[bigint] = mapped_column(
        ForeignKey("users.user_id"), init=False, nullable=False, index=True
    )
    payload: Mapped[json_] = mapped_column(nullable=False)
    created_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)
    last_modified_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)

    # Relationships
    target_owner: Mapped["Group"] = relationship(
        back_populates="draft_resources", lazy="joined"
    )
    creator: Mapped["User"] = relationship(
        back_populates="created_drafts", lazy="joined"
    )

    # Additional settings
    __table_args__ = (Index(None, "group_id", "resource_type", "user_id"),)

    # Initialize default values using dataclass __post_init__ method
    # https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__
    def __post_init__(self) -> None:
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self.created_on = timestamp
        self.last_modified_on = timestamp


class ResourceSnapshot(db.Model):  # type: ignore[name-defined]
    __tablename__ = "resource_snapshots"

    # Database fields
    resource_snapshot_id: Mapped[intpk] = mapped_column(init=False)
    resource_id: Mapped[bigint] = mapped_column(init=False, nullable=False)
    resource_type: Mapped[text_] = mapped_column(
        ForeignKey("resource_types.resource_type"),
        init=False,
        nullable=False,
        index=True,
    )
    user_id: Mapped[bigint] = mapped_column(
        ForeignKey("users.user_id"), init=False, nullable=False
    )
    description: Mapped[optionalstr] = mapped_column(index=True)
    created_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)

    # Relationships
    resource: Mapped["Resource"] = relationship(
        back_populates="versions", lazy="joined"
    )
    creator: Mapped["User"] = relationship(
        back_populates="created_snapshots", lazy="joined"
    )

    # Proxy relationships
    parents: AssociationProxy[list["Resource"]] = association_proxy(
        "resource", "parents", init=False
    )
    children: AssociationProxy[list["Resource"]] = association_proxy(
        "resource", "children", init=False
    )
    tags: AssociationProxy[list["Resource"]] = association_proxy(
        "resource", "tags", init=False
    )

    # Additional settings
    __table_args__ = (
        Index(
            None, "resource_snapshot_id", "resource_id", "resource_type", unique=True
        ),
        Index(None, "resource_id", "created_on"),
        ForeignKeyConstraint(
            ["resource_id", "resource_type"],
            ["resources.resource_id", "resources.resource_type"],
        ),
        UniqueConstraint("resource_snapshot_id", "resource_id"),
    )
    __mapper_args__ = {
        "polymorphic_identity": "resource_snapshot",
        "polymorphic_on": "resource_type",
    }

    # Initialize default values using dataclass __post_init__ method
    # https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__
    def __post_init__(self) -> None:
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self.created_on = timestamp


class Resource(db.Model):  # type: ignore[name-defined]
    __tablename__ = "resources"

    # Database fields
    resource_id: Mapped[intpk] = mapped_column(init=False)
    group_id: Mapped[bigint] = mapped_column(
        ForeignKey("groups.group_id"), init=False, nullable=False
    )
    resource_type: Mapped[text_] = mapped_column(
        ForeignKey("resource_types.resource_type"), nullable=False, index=True
    )
    created_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)

    # Derived fields (read-only)
    last_modified_on: Mapped[datetimetz] = column_property(
        select(func.max(ResourceSnapshot.created_on))
        .where(ResourceSnapshot.resource_id == resource_id)
        .correlate_except(ResourceSnapshot)
        .scalar_subquery()
    )
    is_deleted: Mapped[bool] = column_property(
        select(ResourceLock.resource_id)
        .where(
            ResourceLock.resource_id == resource_id,
            ResourceLock.resource_lock_type == resource_lock_types.DELETE,
        )
        .correlate_except(ResourceLock)
        .exists()
    )
    is_readonly: Mapped[bool] = column_property(
        select(ResourceLock.resource_id)
        .where(
            ResourceLock.resource_id == resource_id,
            ResourceLock.resource_lock_type == resource_lock_types.READONLY,
        )
        .correlate_except(ResourceLock)
        .exists()
    )
    latest_snapshot_id: Mapped[optionalbigint] = column_property(
        Nullable(
            select(ResourceSnapshot.resource_snapshot_id)
            .where(ResourceSnapshot.resource_id == resource_id)
            .order_by(ResourceSnapshot.created_on.desc())
            .limit(1)
            .correlate_except(ResourceSnapshot)
            .scalar_subquery()
        )
    )

    # Relationships
    owner: Mapped["Group"] = relationship(back_populates="resources", lazy="joined")
    parents: Mapped[list["Resource"]] = relationship(
        init=False,
        secondary="resource_dependencies",
        back_populates="children",
        primaryjoin=and_(
            resource_id == resource_dependencies_table.c.child_resource_id,
            resource_type == resource_dependencies_table.c.child_resource_type,
        ),
        secondaryjoin=and_(
            resource_id == resource_dependencies_table.c.parent_resource_id,
            resource_type == resource_dependencies_table.c.parent_resource_type,
        ),
    )
    children: Mapped[list["Resource"]] = relationship(
        init=False,
        secondary="resource_dependencies",
        back_populates="parents",
        primaryjoin=and_(
            resource_id == resource_dependencies_table.c.parent_resource_id,
            resource_type == resource_dependencies_table.c.parent_resource_type,
        ),
        secondaryjoin=and_(
            resource_id == resource_dependencies_table.c.child_resource_id,
            resource_type == resource_dependencies_table.c.child_resource_type,
        ),
    )
    locks: Mapped[list["ResourceLock"]] = relationship(
        init=False, back_populates="resource"
    )
    shares: Mapped[list["SharedResource"]] = relationship(
        init=False, back_populates="resource"
    )
    versions: Mapped[list["ResourceSnapshot"]] = relationship(
        init=False, back_populates="resource"
    )
    tags: Mapped[list["Tag"]] = relationship(
        init=False, secondary="resource_tags", back_populates="resources"
    )

    # Additional settings
    __table_args__ = (
        Index(None, "resource_id", "resource_type", unique=True),
        Index(None, "group_id", "resource_type"),
    )

    # Initialize default values using dataclass __post_init__ method
    # https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__
    def __post_init__(self) -> None:
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self.created_on = timestamp


class SharedResource(db.Model):  # type: ignore[name-defined]
    __tablename__ = "shared_resources"

    # Database fields
    shared_resource_id: Mapped[intpk] = mapped_column(init=False)
    resource_id: Mapped[bigint] = mapped_column(
        ForeignKey("resources.resource_id"), init=False, nullable=False, index=True
    )
    group_id: Mapped[bigint] = mapped_column(
        ForeignKey("groups.group_id"), init=False, nullable=False
    )
    user_id: Mapped[bigint] = mapped_column(
        ForeignKey("users.user_id"), init=False, nullable=False
    )
    read: Mapped[bool] = mapped_column(nullable=False)
    write: Mapped[bool] = mapped_column(nullable=False)
    created_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)

    # Relationships
    resource: Mapped["Resource"] = relationship(back_populates="shares", lazy="joined")
    recipient: Mapped["Group"] = relationship(
        back_populates="received_shares", lazy="joined"
    )
    creator: Mapped["User"] = relationship(
        back_populates="created_shares", lazy="joined"
    )
    tags: Mapped[list["Tag"]] = relationship(
        init=False, secondary="shared_resource_tags", back_populates="shared_resources"
    )

    # Additional settings
    __table_args__ = (Index(None, "group_id", "resource_id"),)

    # Initialize default values using dataclass __post_init__ method
    # https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__
    def __post_init__(self) -> None:
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self.created_on = timestamp


draft_resource_json_resource_id = Index(
    "ix_draft_resources_payload_resource_id",
    DraftResource.payload["resource_id"].as_string().cast(Integer),
)



"""empty message

Revision ID: 7086ad09e884
Revises: 018130a0bf6c
Create Date: 2023-10-09 18:43:21.964011

"""

import datetime
import uuid

import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import sessionmaker

from dioptra.restapi.db.custom_types import GUID

try:
    from sqlalchemy.orm import declarative_base

except ImportError:
    from sqlalchemy.ext.declarative import declarative_base

# Migration data models
UpgradeBase = declarative_base()
DowngradeBase = declarative_base()

# revision identifiers, used by Alembic.
revision = "7086ad09e884"
down_revision = "018130a0bf6c"
branch_labels = None
depends_on = None


class UserUpgrade(UpgradeBase):
    __tablename__ = "users"

    user_id = sa.Column(
        sa.BigInteger().with_variant(sa.Integer, "sqlite"), primary_key=True
    )
    alternative_id = sa.Column(GUID())


class UserDowngrade(DowngradeBase):
    __tablename__ = "users"

    user_id = sa.Column(
        sa.BigInteger().with_variant(sa.Integer, "sqlite"), primary_key=True
    )
    user_expire_on = sa.Column(sa.DateTime())


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    Session = sessionmaker(bind=bind)

    with op.batch_alter_table("users", schema=None) as batch_op:
        batch_op.add_column(sa.Column("alternative_id", GUID()))
        batch_op.drop_column("user_expire_on")

    with Session() as session:
        user_stmt = sa.select(UserUpgrade)

        for user in session.scalars(user_stmt):
            user.alternative_id = uuid.uuid4()

        session.commit()

    with op.batch_alter_table("users", schema=None) as batch_op:
        batch_op.alter_column("alternative_id", nullable=False)
        batch_op.create_unique_constraint(
            batch_op.f("uq_users_alternative_id"), ["alternative_id"]
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    Session = sessionmaker(bind=bind)

    with op.batch_alter_table("users", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("uq_users_alternative_id"), type_="unique")
        batch_op.drop_column("alternative_id")
        batch_op.add_column(sa.Column("user_expire_on", sa.DateTime()))

    with Session() as session:
        user_stmt = sa.select(UserDowngrade)

        for user in session.scalars(user_stmt):
            user.user_expire_on = datetime.datetime(9999, 12, 31, 23, 59, 59)

        session.commit()

    with op.batch_alter_table("users", schema=None) as batch_op:
        batch_op.alter_column("user_expire_on", nullable=False)

    # ### end Alembic commands ###



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from pathlib import Path

from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import select, text

from dioptra.restapi.db import models

# CAUTION!!!
# Using __file__ to get a filepath within the tests/ directory is necessary, but DO NOT
# DO THIS in src/dioptra, use importlib.resources instead!
LATEST_ML_MODELS_SQL_PATH = Path(__file__).parent / "latest_ml_models.sql"


def get_latest_ml_model(db: SQLAlchemy, resource_id: int) -> models.MlModel | None:
    """Get the latest ML model for a given resource ID.

    Args:
        db: The SQLAlchemy database session.
        resource_id: The ID of the resource.

    Returns:
        The latest ML model for the given resource ID, or None if no ML model is found.
    """
    textual_sql = (
        text(LATEST_ML_MODELS_SQL_PATH.read_text())
        .columns(
            models.MlModel.resource_snapshot_id,
            models.MlModel.resource_id,
            models.MlModel.resource_type,
        )
        .bindparams(resource_id=resource_id)
    )
    stmt = select(models.MlModel).from_statement(textual_sql)
    return db.session.execute(stmt).scalar()



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform plugin endpoint operations."""
from __future__ import annotations

import itertools
from typing import Any, Final, cast

import structlog
from flask_login import current_user
from injector import inject
from sqlalchemy import Integer, func, select
from sqlalchemy.orm import aliased
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.db.models.constants import resource_lock_types
from dioptra.restapi.errors import (
    BackendDatabaseError,
    EntityDoesNotExistError,
    EntityExistsError,
    QueryParameterNotUniqueError,
    SortParameterValidationError,
)
from dioptra.restapi.utils import find_non_unique
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.groups.service import GroupIdService
from dioptra.restapi.v1.shared.search_parser import construct_sql_query_filters

LOGGER: BoundLogger = structlog.stdlib.get_logger()

PLUGIN_RESOURCE_TYPE: Final[str] = "plugin"
PLUGIN_FILE_RESOURCE_TYPE: Final[str] = "plugin_file"
PLUGIN_SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "name": lambda x: models.Plugin.name.like(x, escape="/"),
    "description": lambda x: models.Plugin.description.like(x, escape="/"),
}
PLUGIN_FILE_SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "filename": lambda x: models.PluginFile.filename.like(x, escape="/"),
    "description": lambda x: models.PluginFile.description.like(x, escape="/"),
    "contents": lambda x: models.PluginFile.contents.like(x, escape="/"),
}
PLUGIN_SORTABLE_FIELDS: Final[dict[str, Any]] = {
    "name": models.Plugin.name,
    "createdOn": models.Plugin.created_on,
    "lastModifiedOn": models.Resource.last_modified_on,
    "description": models.Plugin.description,
}
PLUGIN_FILE_SORTABLE_FIELDS: Final[dict[str, Any]] = {
    "filename": models.PluginFile.filename,
    "createdOn": models.PluginFile.created_on,
    "lastModifiedOn": models.Resource.last_modified_on,
    "description": models.PluginFile.description,
}


class PluginService(object):
    @inject
    def __init__(
        self, plugin_name_service: PluginNameService, group_id_service: GroupIdService
    ) -> None:
        """Initialize the plugin service.

        All arguments are provided via dependency injection.

        Args:
            plugin_name_service: A PluginNameService object.
            group_id_service: A GroupIdService object.
        """
        self._plugin_name_service = plugin_name_service
        self._group_id_service = group_id_service

    def create(
        self, name: str, description: str, group_id: int, commit: bool = True, **kwargs
    ) -> utils.PluginWithFilesDict:
        """Create a new plugin.

        Args:
            name: The name of the plugin. The combination of name and group_id must be
                unique.
            description: The description of the plugin.
            group_id: The group that will own the plugin.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The newly created plugin object.

        Raises:
            EntityExistsError: If a plugin with the given name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        duplicate = self._plugin_name_service.get(name, group_id=group_id, log=log)
        if duplicate is not None:
            raise EntityExistsError(
                PLUGIN_RESOURCE_TYPE,
                duplicate.resource_id,
                name=name,
                group_id=group_id,
            )

        group = self._group_id_service.get(group_id, error_if_not_found=True)

        resource = models.Resource(resource_type="plugin", owner=group)
        new_plugin = models.Plugin(
            name=name, description=description, resource=resource, creator=current_user
        )
        db.session.add(new_plugin)

        if commit:
            db.session.commit()
            log.debug(
                "Plugin registration successful",
                plugin_id=new_plugin.resource_id,
                name=new_plugin.name,
            )

        plugin_dict = utils.PluginWithFilesDict(
            plugin=new_plugin, plugin_files=[], has_draft=False
        )
        return plugin_dict

    def get(
        self,
        group_id: int | None,
        search_string: str,
        page_index: int,
        page_length: int,
        sort_by_string: str,
        descending: bool,
        **kwargs,
    ) -> tuple[list[utils.PluginWithFilesDict], int]:
        """Fetch a list of plugins, optionally filtering by search string and paging
        parameters.

        Args:
            group_id: A group ID used to filter results.
            search_string: A search string used to filter results.
            page_index: The index of the first group to be returned.
            page_length: The maximum number of plugins to be returned.
            sort_by_string: The name of the column to sort.
            descending: Boolean indicating whether to sort by descending or not.

        Returns:
            A tuple containing a list of plugins and the total number of plugins
            matching the query.

        Raises:
            SearchNotImplementedError: If a search string is provided.
            BackendDatabaseError: If the database query returns a None when counting
                the number of plugins.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get full list of plugins")

        filters = list()

        if group_id is not None:
            filters.append(models.Resource.group_id == group_id)

        if search_string:
            filters.append(
                construct_sql_query_filters(search_string, PLUGIN_SEARCHABLE_FIELDS)
            )

        stmt = (
            select(func.count(models.Plugin.resource_id))
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.Plugin.resource_snapshot_id,
            )
        )
        total_num_plugins = db.session.scalar(stmt)

        if total_num_plugins is None:
            log.error(
                "The database query returned a None when counting the number of "
                "groups when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_plugins == 0:
            return [], total_num_plugins

        # get latest plugin snapshots
        latest_plugins_stmt = (
            select(models.Plugin)
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.Plugin.resource_snapshot_id,
            )
            .offset(page_index)
            .limit(page_length)
        )

        if sort_by_string and sort_by_string in PLUGIN_SORTABLE_FIELDS:
            sort_column = PLUGIN_SORTABLE_FIELDS[sort_by_string]
            if descending:
                sort_column = sort_column.desc()
            else:
                sort_column = sort_column.asc()
            latest_plugins_stmt = latest_plugins_stmt.order_by(sort_column)
        elif sort_by_string and sort_by_string not in PLUGIN_SORTABLE_FIELDS:
            raise SortParameterValidationError(PLUGIN_RESOURCE_TYPE, sort_by_string)

        plugins = db.session.scalars(latest_plugins_stmt).all()

        # extract list of plugin ids
        plugin_ids = [plugin.resource_id for plugin in plugins]

        # Build CTE that retrieves all snapshot ids for the list of plugin files
        # associated with retrieved plugins
        parent_plugin = aliased(models.Plugin)
        plugin_file_snapshot_ids_cte = (
            select(models.PluginFile.resource_snapshot_id)
            .join(
                models.resource_dependencies_table,
                models.PluginFile.resource_id
                == models.resource_dependencies_table.c.child_resource_id,
            )
            .join(
                parent_plugin,
                parent_plugin.resource_id
                == models.resource_dependencies_table.c.parent_resource_id,
            )
            .where(parent_plugin.resource_id.in_(plugin_ids))
            .cte()
        )

        # get the latest plugin file snapshots associated with the retrieved plugins
        latest_plugin_files_stmt = (
            select(models.PluginFile)
            .join(models.Resource)
            .where(
                models.PluginFile.resource_snapshot_id.in_(
                    select(plugin_file_snapshot_ids_cte)
                ),
                models.Resource.latest_snapshot_id
                == models.PluginFile.resource_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        plugin_files = db.session.scalars(latest_plugin_files_stmt).unique().all()

        # build a dictionary structure to re-associate plugins and plugin files
        plugins_dict: dict[int, utils.PluginWithFilesDict] = {
            plugin.resource_id: utils.PluginWithFilesDict(
                plugin=plugin, plugin_files=[], has_draft=False
            )
            for plugin in plugins
        }
        for plugin_file in plugin_files:
            plugins_dict[plugin_file.plugin_id]["plugin_files"].append(plugin_file)

        drafts_stmt = select(
            models.DraftResource.payload["resource_id"].as_string().cast(Integer)
        ).where(
            models.DraftResource.payload["resource_id"]
            .as_string()
            .cast(Integer)
            .in_(
                tuple(plugin["plugin"].resource_id for plugin in plugins_dict.values())
            ),
            models.DraftResource.user_id == current_user.user_id,
        )
        for resource_id in db.session.scalars(drafts_stmt):
            plugins_dict[resource_id]["has_draft"] = True

        return list(plugins_dict.values()), total_num_plugins


class PluginIdService(object):
    """The service methods for registering and managing plugins by their unique id."""

    @inject
    def __init__(
        self,
        plugin_name_service: PluginNameService,
    ) -> None:
        """Initialize the plugin service.

        All arguments are provided via dependency injection.

        Args:
            plugin_name_service: A pluginNameService object.
        """
        self._plugin_name_service = plugin_name_service

    def get(
        self,
        plugin_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> utils.PluginWithFilesDict | None:
        """Fetch a plugin by its unique id.

        Args:
            plugin_id: The unique id of the plugin.
            error_if_not_found: If True, raise an error if the plugin is not found.
                Defaults to False.

        Returns:
            The plugin object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the plugin is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get plugin by id", plugin_id=plugin_id)

        stmt = (
            select(models.Plugin)
            .join(models.Resource)
            .where(
                models.Plugin.resource_id == plugin_id,
                models.Plugin.resource_snapshot_id
                == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        plugin = db.session.scalar(stmt)

        if plugin is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(PLUGIN_RESOURCE_TYPE, plugin_id=plugin_id)

            return None

        latest_plugin_files_stmt = (
            select(models.PluginFile)
            .join(models.Resource)
            .where(
                models.PluginFile.plugin_id == plugin_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.PluginFile.resource_snapshot_id,
            )
        )
        plugin_files = list(db.session.scalars(latest_plugin_files_stmt).unique().all())

        drafts_stmt = (
            select(models.DraftResource.draft_resource_id)
            .where(
                models.DraftResource.payload["resource_id"].as_string().cast(Integer)
                == plugin.resource_id,
                models.DraftResource.user_id == current_user.user_id,
            )
            .exists()
            .select()
        )
        has_draft = db.session.scalar(drafts_stmt)

        return utils.PluginWithFilesDict(
            plugin=plugin, plugin_files=plugin_files, has_draft=has_draft
        )

    def modify(
        self,
        plugin_id: int,
        name: str,
        description: str,
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> utils.PluginWithFilesDict | None:
        """Rename a plugin.

        Args:
            plugin_id: The unique id of the plugin.
            name: The new name of the plugin.
            description: The new description of the plugin.
            error_if_not_found: If True, raise an error if the plugin is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated plugin object.

        Raises:
            EntityDoesNotExistError: If the plugin is not found and `error_if_not_found`
                is True.
            EntityExistsError: If the plugin name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        plugin_dict = self.get(
            plugin_id, error_if_not_found=error_if_not_found, log=log
        )

        if plugin_dict is None:
            return None

        plugin = plugin_dict["plugin"]
        plugin_files = plugin_dict["plugin_files"]
        group_id = plugin.resource.group_id

        if name != plugin.name:
            duplicate = self._plugin_name_service.get(name, group_id=group_id, log=log)
            if duplicate is not None:
                raise EntityExistsError(
                    PLUGIN_RESOURCE_TYPE,
                    duplicate.resource_id,
                    name=name,
                    group_id=group_id,
                )

        new_plugin = models.Plugin(
            name=name,
            description=description,
            resource=plugin.resource,
            creator=current_user,
        )
        db.session.add(new_plugin)

        if commit:
            db.session.commit()
            log.debug(
                "Plugin modification successful",
                plugin_id=plugin_id,
                name=name,
                description=description,
            )

        return utils.PluginWithFilesDict(
            plugin=new_plugin, plugin_files=plugin_files, has_draft=False
        )

    def delete(self, plugin_id: int, **kwargs) -> dict[str, Any]:
        """Delete a plugin.

        Args:
            plugin_id: The unique id of the plugin.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        stmt = select(models.Resource).filter_by(
            resource_id=plugin_id, resource_type=PLUGIN_RESOURCE_TYPE, is_deleted=False
        )
        plugin_resource = db.session.scalar(stmt)

        if plugin_resource is None:
            raise EntityDoesNotExistError(PLUGIN_RESOURCE_TYPE, plugin_id=plugin_id)

        deleted_resource_lock = models.ResourceLock(
            resource_lock_type=resource_lock_types.DELETE,
            resource=plugin_resource,
        )
        db.session.add(deleted_resource_lock)
        db.session.commit()
        log.debug("Plugin deleted", plugin_id=plugin_id)

        return {"status": "Success", "id": [plugin_id]}


class PluginIdsService(object):
    """The service methods for registering and managing plugins by their unique id."""

    def get(
        self,
        plugin_ids: list[int],
        error_if_not_found: bool = False,
        **kwargs,
    ) -> list[utils.PluginWithFilesDict]:
        """Fetch a plugin by its unique id.

        Args:
            plugin_ids: The unique ids of the plugins.
            error_if_not_found: If True, raise an error if the plugin is not found.
                Defaults to False.

        Returns:
            A list of  plugin objects.

        Raises:
            EntityDoesNotExistError: If the plugin is not found and `error_if_not_found`
                is True.
        """
        latest_plugins_stmt = (
            select(models.Plugin)
            .join(models.Resource)
            .where(
                models.Plugin.resource_id.in_(tuple(plugin_ids)),
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.Plugin.resource_snapshot_id,
            )
        )
        plugins = db.session.scalars(latest_plugins_stmt).all()

        if len(plugins) != len(plugin_ids) and error_if_not_found:
            plugin_ids_missing = set(plugin_ids) - set(
                plugin.resource_id for plugin in plugins
            )
            raise EntityDoesNotExistError(
                PLUGIN_RESOURCE_TYPE, plugin_ids=list(plugin_ids_missing)
            )

        # extract list of plugin ids
        plugin_ids = [plugin.resource_id for plugin in plugins]

        # Build CTE that retrieves all snapshot ids for the list of plugin files
        # associated with retrieved plugins
        parent_plugin = aliased(models.Plugin)
        plugin_file_snapshot_ids_cte = (
            select(models.PluginFile.resource_snapshot_id)
            .join(
                models.resource_dependencies_table,
                models.PluginFile.resource_id
                == models.resource_dependencies_table.c.child_resource_id,
            )
            .join(
                parent_plugin,
                parent_plugin.resource_id
                == models.resource_dependencies_table.c.parent_resource_id,
            )
            .where(parent_plugin.resource_id.in_(plugin_ids))
            .cte()
        )

        # get the latest plugin file snapshots associated with the retrieved plugins
        latest_plugin_files_stmt = (
            select(models.PluginFile)
            .join(models.Resource)
            .where(
                models.Resource.latest_snapshot_id.in_(
                    select(plugin_file_snapshot_ids_cte)
                ),
                models.Resource.latest_snapshot_id
                == models.PluginFile.resource_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        plugin_files = db.session.scalars(latest_plugin_files_stmt).unique().all()

        # build a dictionary structure to re-associate plugins and plugin files
        plugins_dict: dict[int, utils.PluginWithFilesDict] = {
            plugin.resource_id: utils.PluginWithFilesDict(
                plugin=plugin, plugin_files=[], has_draft=False
            )
            for plugin in plugins
        }
        for plugin_file in plugin_files:
            plugins_dict[plugin_file.plugin_id]["plugin_files"].append(plugin_file)

        drafts_stmt = select(
            models.DraftResource.payload["resource_id"].as_string().cast(Integer)
        ).where(
            models.DraftResource.payload["resource_id"]
            .as_string()
            .cast(Integer)
            .in_(
                tuple(plugin["plugin"].resource_id for plugin in plugins_dict.values())
            ),
            models.DraftResource.user_id == current_user.user_id,
        )
        for resource_id in db.session.scalars(drafts_stmt):
            plugins_dict[resource_id]["has_draft"] = True

        return list(plugins_dict.values())


class PluginNameService(object):
    """The service methods for managing plugins by their name."""

    def get(
        self,
        name: str,
        group_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> models.Plugin | None:
        """Fetch a plugin by its name.

        Args:
            name: The name of the plugin.
            group_id: The the group id of the plugin.
            error_if_not_found: If True, raise an error if the plugin is not found.
                Defaults to False.

        Returns:
            The plugin object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the plugin is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get plugin by name", plugin_name=name, group_id=group_id)

        stmt = (
            select(models.Plugin)
            .join(models.Resource)
            .where(
                models.Plugin.name == name,
                models.Resource.group_id == group_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.Plugin.resource_snapshot_id,
            )
        )
        plugin = db.session.scalar(stmt)

        if plugin is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    PLUGIN_RESOURCE_TYPE, name=name, group_id=group_id
                )

        return plugin


class PluginFileNameService(object):
    """The service methods for managing plugin files by their filename."""

    def get(
        self, filename: str, plugin_id: int, error_if_not_found: bool = False, **kwargs
    ) -> models.PluginFile | None:
        """Fetch a plugin file by its filename.

        Args:
            filename: The filename of the plugin file.
            plugin_id: The plugin id of the plugin to which the plugin file belongs.
            error_if_not_found: If True, raise an error if the plugin file is not found.
                Defaults to False.

        Returns:
            The plugin file object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the plugin is not found and
            `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug(
            "Get plugin file by filename",
            plugin_file_filename=filename,
            plugin_id=plugin_id,
        )

        stmt = (
            select(models.PluginFile)
            .join(models.Resource)
            .where(
                models.PluginFile.filename == filename,
                models.PluginFile.plugin_id == plugin_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.PluginFile.resource_snapshot_id,
            )
        )
        plugin_file = db.session.scalar(stmt)

        if plugin_file is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    PLUGIN_FILE_RESOURCE_TYPE, plugin_id=plugin_id, filename=filename
                )

            return None

        return plugin_file


class PluginIdFileService(object):
    @inject
    def __init__(
        self,
        plugin_file_name_service: PluginFileNameService,
        plugin_id_service: PluginIdService,
        group_id_service: GroupIdService,
    ) -> None:
        """Initialize the plugin file service.

        All arguments are provided via dependency injection.

        Args:
            plugin_file_name_service: A PluginFileNameService object.
            plugin_id_service: A PluginIdService object.
            group_id_service: A GroupIdService object.
        """
        self._plugin_file_name_service = plugin_file_name_service
        self._plugin_id_service = plugin_id_service
        self._group_id_service = group_id_service

    def create(
        self,
        filename: str,
        contents: str,
        description: str,
        tasks: list[dict[str, Any]],
        plugin_id: int,
        commit: bool = True,
        **kwargs,
    ) -> utils.PluginFileDict:
        """Creates a new PluginFile object.

        The PluginFile object will belong to the group that owns the plugin associated
        with the PluginFile. The creator will be the current user.

        Args:
            filename: The name of the plugin file.
            contents: The contents of the plugin file.
            description: The description of the plugin file.
            tasks: The tasks associated with the plugin file.
            plugin_id: The unique id of the plugin containing the plugin file.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The newly created plugin file object.

        Raises:
            EntityExistsError: If a plugin file with the given filename
                already exists.
        """

        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Create a new plugin file", plugin_id=plugin_id, filename=filename)

        # Check if plugin_id points to a Plugin that exists, and if so, retrieve it.
        plugin_dict = cast(
            utils.PluginWithFilesDict,
            self._plugin_id_service.get(plugin_id, error_if_not_found=True),
        )

        # Validate that the proposed filename hasn't already been used in the plugin.
        duplicate = self._plugin_file_name_service.get(
            filename, plugin_id=plugin_id, log=log
        )
        if duplicate is not None:
            raise EntityExistsError(
                PLUGIN_FILE_RESOURCE_TYPE,
                duplicate.resource_id,
                filename=filename,
                plugin_id=plugin_id,
            )

        # The owner of the PluginFile resource must match the owner of the Plugin
        # resource.
        plugin = plugin_dict["plugin"]
        resource = models.Resource(
            resource_type=PLUGIN_FILE_RESOURCE_TYPE, owner=plugin.resource.owner
        )
        new_plugin_file = models.PluginFile(
            filename=filename,
            contents=contents,
            description=description,
            resource=resource,
            creator=current_user,
        )

        new_plugin_file.parents.append(plugin.resource)
        db.session.add(new_plugin_file)

        _add_plugin_tasks(tasks, plugin_file=new_plugin_file, log=log)

        if commit:
            db.session.commit()
            log.debug(
                "Plugin file registration successful",
                plugin_file_id=new_plugin_file.resource_id,
                filename=new_plugin_file.filename,
            )

        return utils.PluginFileDict(
            plugin_file=new_plugin_file, plugin=plugin, has_draft=False
        )

    def get(
        self,
        plugin_id: int,
        search_string: str,
        page_index: int,
        page_length: int,
        sort_by_string: str,
        descending: bool,
        **kwargs,
    ) -> tuple[list[utils.PluginFileDict], int]:
        """Fetch a list of plugin files associated with a plugin, optionally
        filtering by search string and paging parameters.

        Args:
            plugin_id: A plugin ID used to filter results.
            search_string: A search string used to filter results.
            page_index: The index of the first group to be returned.
            page_length: The maximum number of plugins to be returned.
            sort_by_string: The name of the column to sort.
            descending: Boolean indicating whether to sort by descending or not.

        Returns:
            A tuple containing a list of plugins and the total number of plugins
            matching the query.

        Raises:
            SearchNotImplementedError: If a search string is provided.
            BackendDatabaseError: If the database query returns a None when counting
                the number of plugins.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug(
            "Get list of plugin files",
            plugin_id=plugin_id,
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
        )

        filters = list()

        if search_string:
            filters.append(
                construct_sql_query_filters(
                    search_string, PLUGIN_FILE_SEARCHABLE_FIELDS
                )
            )

        latest_plugin_stmt = (
            select(models.Plugin)
            .join(models.Resource)
            .where(
                models.Plugin.resource_id == plugin_id,
                models.Plugin.resource_snapshot_id
                == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        plugin = db.session.scalar(latest_plugin_stmt)

        if plugin is None:
            raise EntityDoesNotExistError(PLUGIN_RESOURCE_TYPE, plugin_id=plugin_id)

        latest_plugin_files_count_stmt = (
            select(func.count(models.PluginFile.resource_id))
            .join(models.Resource)
            .where(
                *filters,
                models.PluginFile.plugin_id == plugin_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.PluginFile.resource_snapshot_id,
            )
        )
        total_num_plugin_files = db.session.scalar(latest_plugin_files_count_stmt)

        if total_num_plugin_files is None:
            log.error(
                "The database query returned a None when counting the number of "
                "groups when it should return a number.",
                sql=str(latest_plugin_files_count_stmt),
            )
            raise BackendDatabaseError

        if total_num_plugin_files == 0:
            return [], total_num_plugin_files

        latest_plugin_files_stmt = (
            select(models.PluginFile)
            .join(models.Resource)
            .where(
                *filters,
                models.PluginFile.plugin_id == plugin_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.PluginFile.resource_snapshot_id,
            )
            .offset(page_index)
            .limit(page_length)
        )

        if sort_by_string and sort_by_string in PLUGIN_FILE_SORTABLE_FIELDS:
            sort_column = PLUGIN_FILE_SORTABLE_FIELDS[sort_by_string]
            if descending:
                sort_column = sort_column.desc()
            else:
                sort_column = sort_column.asc()
            latest_plugin_files_stmt = latest_plugin_files_stmt.order_by(sort_column)
        elif sort_by_string and sort_by_string not in PLUGIN_FILE_SORTABLE_FIELDS:
            raise SortParameterValidationError(
                PLUGIN_FILE_RESOURCE_TYPE, sort_by_string
            )

        plugin_files_dict: dict[int, utils.PluginFileDict] = {
            plugin_file.resource_id: utils.PluginFileDict(
                plugin=plugin, plugin_file=plugin_file, has_draft=False
            )
            for plugin_file in db.session.scalars(latest_plugin_files_stmt).unique()
        }

        drafts_stmt = select(
            models.DraftResource.payload["resource_id"].as_string().cast(Integer)
        ).where(
            models.DraftResource.payload["resource_id"]
            .as_string()
            .cast(Integer)
            .in_(
                tuple(
                    plugin_file["plugin_file"].resource_id
                    for plugin_file in plugin_files_dict.values()
                )
            ),
            models.DraftResource.user_id == current_user.user_id,
        )
        for resource_id in db.session.scalars(drafts_stmt):
            plugin_files_dict[resource_id]["has_draft"] = True

        return list(plugin_files_dict.values()), total_num_plugin_files

    def delete(self, plugin_id: int, **kwargs) -> dict[str, Any]:
        """Deletes all plugin files associated with a plugin.

        Args:
            plugin_id: The unique id of the plugin.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        stmt = select(models.Resource).filter_by(
            resource_id=plugin_id, resource_type=PLUGIN_RESOURCE_TYPE, is_deleted=False
        )

        plugin_resource = db.session.scalar(stmt)

        if plugin_resource is None:
            raise EntityDoesNotExistError(PLUGIN_RESOURCE_TYPE, plugin_id=plugin_id)

        latest_plugin_files_stmt = (
            select(models.PluginFile)
            .join(models.Resource)
            .where(
                models.PluginFile.plugin_id == plugin_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.PluginFile.resource_snapshot_id,
            )
        )

        plugin_files = db.session.scalars(latest_plugin_files_stmt).unique().all()
        num_plugin_files = len(plugin_files)

        plugin_file_ids = []
        for plugin_file in plugin_files:
            db.session.add(
                models.ResourceLock(
                    resource_lock_type=resource_lock_types.DELETE,
                    resource=plugin_file.resource,
                )
            )
            plugin_file_ids.append(plugin_file.resource_id)

        db.session.commit()

        log.debug(
            "Plugin files deleted",
            plugin_id=plugin_id,
            num_plugin_files=num_plugin_files,
        )

        return {"status": "Success", "id": plugin_file_ids}


class PluginIdFileIdService(object):
    @inject
    def __init__(self, plugin_file_name_service: PluginFileNameService):
        self._plugin_file_name_service = plugin_file_name_service

    def get(
        self,
        plugin_id: int,
        plugin_file_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> utils.PluginFileDict | None:
        """Fetch a plugin file by its unique id.

        Args:
            plugin_id: The unique id of the plugin containing the plugin file.
            plugin_file_id: The unique id of the plugin file.
            error_if_not_found: If True, raise an error if the plugin or plugin file is
                not found. Defaults to False.

        Returns:
            The plugin object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the plugin or plugin file is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug(
            "Get plugin file by id", plugin_id=plugin_id, plugin_file_id=plugin_file_id
        )

        latest_plugin_stmt = (
            select(models.Plugin)
            .join(models.Resource)
            .where(
                models.Plugin.resource_id == plugin_id,
                models.Plugin.resource_snapshot_id
                == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        plugin = db.session.scalar(latest_plugin_stmt)

        if plugin is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(PLUGIN_RESOURCE_TYPE, plugin_id=plugin_id)

            return None

        latest_plugin_file_stmt = (
            select(models.PluginFile)
            .join(models.Resource)
            .where(
                models.PluginFile.plugin_id == plugin_id,
                models.PluginFile.resource_id == plugin_file_id,
                models.PluginFile.resource_snapshot_id
                == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        plugin_file = db.session.scalar(latest_plugin_file_stmt)

        if plugin_file is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    PLUGIN_FILE_RESOURCE_TYPE,
                    plugin_id=plugin_id,
                    plugin_file_id=plugin_file_id,
                )

            return None

        drafts_stmt = (
            select(models.DraftResource.draft_resource_id)
            .where(
                models.DraftResource.payload["resource_id"].as_string().cast(Integer)
                == plugin_file.resource_id,
                models.DraftResource.user_id == current_user.user_id,
            )
            .exists()
            .select()
        )
        has_draft = db.session.scalar(drafts_stmt)

        return utils.PluginFileDict(
            plugin_file=plugin_file, plugin=plugin, has_draft=has_draft
        )

    def modify(
        self,
        plugin_id: int,
        plugin_file_id: int,
        filename: str,
        contents: str,
        description: str,
        tasks: list,
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> utils.PluginFileDict | None:
        """Modify a plugin file.

        Args:
            plugin_id: The unique id of the plugin.
            plugin_file_id: The unique id of the plugin file.
            filename: The updated filename of the plugin file.
            contents: The updated contents of the plugin file.
            description: The updated description of the plugin file.
            tasks: The updated tasks associated with the plugin file.
            error_if_not_found: If True, raise an error if the plugin or plugin file is
                not found. Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated plugin file object.

        Raises:
            EntityDoesNotExistError: If the plugin or plugin file is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        plugin_file_dict = self.get(
            plugin_id=plugin_id,
            plugin_file_id=plugin_file_id,
            error_if_not_found=error_if_not_found,
            log=log,
        )

        if plugin_file_dict is None:
            return None

        plugin = plugin_file_dict["plugin"]
        plugin_file = plugin_file_dict["plugin_file"]

        if filename != plugin_file.filename:
            duplicate = self._plugin_file_name_service.get(
                filename, plugin_id=plugin_id, log=log
            )
            if duplicate is not None:
                raise EntityExistsError(
                    PLUGIN_FILE_RESOURCE_TYPE,
                    duplicate.resource_id,
                    filename=filename,
                    plugin_id=plugin_id,
                )

        updated_plugin_file = models.PluginFile(
            filename=filename,
            contents=contents,
            description=description,
            resource=plugin_file.resource,
            creator=current_user,
        )
        db.session.add(updated_plugin_file)

        _add_plugin_tasks(tasks, plugin_file=updated_plugin_file, log=log)

        if commit:
            db.session.commit()
            log.debug(
                "Plugin file modification successful",
                plugin_file_id=updated_plugin_file.resource_id,
                filename=updated_plugin_file.filename,
            )

        return utils.PluginFileDict(
            plugin_file=updated_plugin_file, plugin=plugin, has_draft=False
        )

    def delete(self, plugin_id: int, plugin_file_id: int, **kwargs) -> dict[str, Any]:
        """Deletes a plugin file.

        Args:
            plugin_id: The unique id of the plugin.
            plugin_file_id: The unique id of the plugin file.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        stmt = select(models.Resource).filter_by(
            resource_id=plugin_id, resource_type=PLUGIN_RESOURCE_TYPE, is_deleted=False
        )

        plugin_resource = db.session.scalar(stmt)
        if plugin_resource is None:
            raise EntityDoesNotExistError(PLUGIN_RESOURCE_TYPE, plugin_id=plugin_id)

        plugin_file_stmt = (
            select(models.PluginFile)
            .join(models.Resource)
            .where(
                models.PluginFile.plugin_id == plugin_id,
                models.PluginFile.resource_id == plugin_file_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.PluginFile.resource_snapshot_id,
            )
        )
        plugin_file = db.session.scalar(plugin_file_stmt)

        if plugin_file is None:
            raise EntityDoesNotExistError(
                PLUGIN_FILE_RESOURCE_TYPE,
                plugin_id=plugin_id,
                plugin_file_id=plugin_file_id,
            )

        plugin_file_id_to_return = plugin_file.resource_id  # to return to user
        db.session.add(
            models.ResourceLock(
                resource_lock_type=resource_lock_types.DELETE,
                resource=plugin_file.resource,
            )
        )
        db.session.commit()

        log.debug(
            "Plugin file deleted",
            plugin_id=plugin_id,
        )

        return {"status": "Success", "id": [plugin_file_id_to_return]}


def _construct_plugin_task(
    plugin_file: models.PluginFile,
    task: dict[str, Any],
    parameter_types_id_to_orm: dict[int, models.PluginTaskParameterType],
    log: BoundLogger,
) -> models.PluginTask:
    duplicates = find_non_unique("name", task["input_params"])
    if len(duplicates) > 0:
        raise QueryParameterNotUniqueError(
            "plugin task input parameter",
            plugin_task_name=task["name"],
            input_param_names=duplicates,
        )

    duplicates = find_non_unique("name", task["output_params"])
    if len(duplicates) > 0:
        raise QueryParameterNotUniqueError(
            "plugin task output parameter",
            plugin_task_name=task["name"],
            output_param_names=duplicates,
        )

    input_parameters_list = []
    for parameter_number, input_param in enumerate(task["input_params"]):
        parameter_type = parameter_types_id_to_orm[input_param["parameter_type_id"]]
        input_parameters_list.append(
            models.PluginTaskInputParameter(
                name=input_param["name"],
                parameter_number=parameter_number,
                parameter_type=parameter_type,
                required=input_param["required"],
            )
        )

    output_parameters_list = []
    for parameter_number, output_param in enumerate(task["output_params"]):
        parameter_type = parameter_types_id_to_orm[output_param["parameter_type_id"]]
        output_parameters_list.append(
            models.PluginTaskOutputParameter(
                name=output_param["name"],
                parameter_number=parameter_number,
                parameter_type=parameter_type,
            )
        )

    plugin_task_orm = models.PluginTask(
        file=plugin_file,
        plugin_task_name=task["name"],
        input_parameters=input_parameters_list,
        output_parameters=output_parameters_list,
    )
    return plugin_task_orm


def _get_referenced_parameter_types(
    tasks: list[dict[str, Any]], log: BoundLogger
) -> dict[int, models.PluginTaskParameterType] | None:
    parameter_type_ids = set(
        [
            param["parameter_type_id"]
            for task in tasks
            for param in itertools.chain(task["input_params"], task["output_params"])
        ]
    )

    if not len(parameter_type_ids) > 0:
        return None

    parameter_types_stmt = (
        select(models.PluginTaskParameterType)
        .join(models.Resource)
        .where(
            models.PluginTaskParameterType.resource_id.in_(tuple(parameter_type_ids)),
            models.Resource.is_deleted == False,  # noqa: E712
            models.Resource.latest_snapshot_id
            == models.PluginTaskParameterType.resource_snapshot_id,
        )
    )
    parameter_types = db.session.scalars(parameter_types_stmt).all()

    if parameter_types is None:
        log.error(
            "The database query returned a None when it should return plugin "
            "parameter types.",
            sql=str(parameter_types_stmt),
            num_expected=len(parameter_type_ids),
        )
        raise BackendDatabaseError

    if not len(parameter_types) == len(parameter_type_ids):
        returned_parameter_type_ids = set([x.resource_id for x in parameter_types])
        ids_not_found = parameter_type_ids - returned_parameter_type_ids
        raise EntityDoesNotExistError(
            "plugin task parameter types",
            num_expected=len(parameter_type_ids),
            num_found=len(parameter_types),
            ids_not_found=sorted(list(ids_not_found)),
        )

    return {x.resource_id: x for x in parameter_types}


def _add_plugin_tasks(
    tasks: list[dict[str, Any]], plugin_file: models.PluginFile, log: BoundLogger
) -> None:
    if not tasks:
        return None

    duplicates = find_non_unique("name", tasks)
    if len(duplicates) > 0:
        raise QueryParameterNotUniqueError("plugin task", task_names=duplicates)

    parameter_types_id_to_orm = _get_referenced_parameter_types(tasks, log=log) or {}
    for task in tasks:
        plugin_task = _construct_plugin_task(
            plugin_file,
            task=task,
            parameter_types_id_to_orm=parameter_types_id_to_orm,
            log=log,
        )
        db.session.add(plugin_task)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""A context manager for a Flask test server.

This module provides a context manager for a Flask test server. This can be used in
combination with a requests Session as an alternative to the Flask test client for
testing the REST API.
"""
from __future__ import annotations

import os
import platform
import shutil
import signal
import subprocess
import time
from pathlib import Path
from subprocess import PIPE, CompletedProcess, Popen


class FlaskTestServer(object):
    """A context manager for a Flask test server.

    This class provides a convenient way to start and stop a Flask server process
    for testing purposes. The server process runs in a subprocess and uses an SQLite
    database for storage.
    """

    def __init__(
        self, sqlite_path: str | Path, extra_env: dict[str, str] | None = None
    ) -> None:
        """Initialize the FlaskTestServer object.

        Args:
            sqlite_path: The path to the SQLite database file.
            extra_env: A dictionary of additional environment variables to set for the
                server process.
        """
        self._process: Popen | None = None
        self._sqlite_uri: str = f"sqlite:///{Path(sqlite_path).resolve()}"
        self._extra_env: dict[str, str] = extra_env or {}

    def __enter__(self) -> FlaskTestServer:
        """Start the Flask server process and enter the runtime context.

        Returns:
            The FlaskTestServer object.
        """
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Stop the Flask server process and exit the runtime context.

        Args:
            exc_type: The type of the exception raised in the runtime context. None if
                no exception was raised.
            exc_value: The value of the exception raised in the runtime context. None
                if no exception was raised.
            traceback: The traceback of the exception raised in the runtime context.
                None if no exception was raised.
        """
        self.stop()

    @property
    def env(self) -> dict[str, str]:
        """A dictionary containing the server's environment variables."""
        return (
            os.environ
            | {
                "DIOPTRA_RESTAPI_DEV_DATABASE_URI": self._sqlite_uri,
                "DIOPTRA_RESTAPI_ENV": "dev",
            }
            | self._extra_env
        )

    @property
    def dioptra_db_cmd(self) -> str:
        """The path to the dioptra-db command.

        Raises:
            RuntimeError: If the dioptra-db command is not found.
        """
        cmd = shutil.which("dioptra-db")

        if cmd is None:
            raise RuntimeError("dioptra-db command not found.")

        return cmd

    @property
    def flask_cmd(self) -> str:
        """The path to the Flask command.

        Raises:
            RuntimeError: If the Flask command is not found.
        """
        cmd = shutil.which("flask")

        if cmd is None:
            raise RuntimeError("Flask command not found.")

        return cmd

    def start(self) -> Popen:
        """Start the Flask server process.

        Returns:
            The Flask server process.
        """
        self._process = Popen(
            args=[self.flask_cmd, "run", "--port", "5000", "--no-reload"],
            env=self.env,
            stdout=PIPE,
            stderr=PIPE,
        )
        return self._process

    def stop(self) -> int | None:
        """Stop the Flask server process.

        Returns:
            The return code of the Flask server process. If the process is not running,
            returns None.
        """
        if self._process is None:
            return None

        stop_attempts = 0
        while (return_code := self._process.poll()) is None:
            if stop_attempts <= 2:
                self._send_interrupt_signal()

            elif stop_attempts <= 4:
                self._process.terminate()

            else:
                self._process.kill()

            stop_attempts += 1
            time.sleep(1)

        self._process = None
        return return_code

    def upgrade_db(self, *args) -> CompletedProcess:
        """Run the Flask database upgrade command in a subprocess.

        Args:
            *args: Additional arguments to pass to the Flask database upgrade command.

        Returns:
            The result of the database upgrade command.
        """
        return subprocess.run(
            args=[self.dioptra_db_cmd, "autoupgrade", *args],
            env=self.env,
            capture_output=True,
            text=True,
        )

    def _send_interrupt_signal(self) -> None:
        """Send OS-dependent interrupt signal to the Flask server process."""
        if platform.system() == "Windows":
            self._process.send_signal(signal.CTRL_C_EVENT)

        if platform.system() in {"Linux", "Darwin"}:
            self._process.send_signal(signal.SIGINT)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""
Build a "type registry", i.e. a mapping from type name to Type instance,
from a set of type definitions.
"""
import graphlib
from collections.abc import Iterable, Iterator, Mapping, MutableMapping, Sequence
from typing import Any, Optional, Union, cast

from dioptra.sdk.exceptions.task_engine import (
    AnonymousSimpleTypeError,
    BuiltinTypeRedefinitionError,
    DioptraTypeError,
    InvalidKeyTypeError,
    NonSimpleSuperTypeError,
    TooManyTypeStructuresError,
    TypeNotFoundError,
    TypeReferenceCycleError,
)
from dioptra.task_engine import types

TYPE_ANY = types.SimpleType("any")
TYPE_NUMBER = types.SimpleType("number")
TYPE_INTEGER = types.SimpleType("integer", super_type=TYPE_NUMBER)
TYPE_STRING = types.SimpleType("string")
TYPE_BOOLEAN = types.SimpleType("boolean")
TYPE_NULL = types.SimpleType("null")


BUILTIN_TYPES = {
    TYPE_ANY.name: TYPE_ANY,
    TYPE_STRING.name: TYPE_STRING,
    TYPE_INTEGER.name: TYPE_INTEGER,
    TYPE_NUMBER.name: TYPE_NUMBER,
    TYPE_BOOLEAN.name: TYPE_BOOLEAN,
    TYPE_NULL.name: TYPE_NULL,
}


# Type aliases for type annotations
_TypeRegistry = Mapping[str, types.Type]
_TypeDefinition = Mapping[str, Any]
_NestedTypeDefOrRef = Union[_TypeDefinition, str]


def _build_mapping_structure(
    map_structure: Union[
        Mapping[str, _NestedTypeDefOrRef], Sequence[_NestedTypeDefOrRef]
    ],
    type_registry: _TypeRegistry,
):
    """
    Build a TypeStructure instance from a definition of a mapping structure.

    Args:
        map_structure: The structure definition
        type_registry: A type registry, used to look up type references

    Returns:
        A mapping structure
    """
    if isinstance(map_structure, Mapping):
        prop_dict = {}
        for prop_name, prop_def in map_structure.items():
            prop_type = build_or_get_type(prop_def, type_registry)
            prop_dict[prop_name] = prop_type

        type_structure = types.TypeStructure(types.StructureType.MAPPING, prop_dict)

    else:
        # A sequence

        # Requires a key type and a value type.
        assert len(map_structure) == 2

        key_type_def = map_structure[0]
        value_type_def = map_structure[1]

        key_type = build_or_get_type(key_type_def, type_registry)
        value_type = build_or_get_type(value_type_def, type_registry)

        if key_type not in (TYPE_STRING, TYPE_INTEGER):
            raise InvalidKeyTypeError(key_type)

        type_structure = types.TypeStructure(
            types.StructureType.MAPPING, [key_type, value_type]
        )

    return type_structure


def _build_list_structure(
    list_structure: _NestedTypeDefOrRef, type_registry: _TypeRegistry
) -> types.TypeStructure:
    """
    Build a TypeStructure instance from a definition of a list structure.

    Args:
        list_structure: The structure definition
        type_registry: A type registry, used to look up type references

    Returns:
        A list structure
    """
    elt_type = build_or_get_type(list_structure, type_registry)

    type_structure = types.TypeStructure(types.StructureType.LIST, elt_type)

    return type_structure


def _build_tuple_structure(
    tuple_structure: Iterable[_NestedTypeDefOrRef], type_registry: _TypeRegistry
) -> types.TypeStructure:
    """
    Build a TypeStructure instance from a definition of a tuple structure.

    Args:
        tuple_structure: The structure definition
        type_registry: A type registry, used to look up type references

    Returns:
        A tuple structure
    """
    elt_types = []
    for elt_structure_def in tuple_structure:
        elt_type = build_or_get_type(elt_structure_def, type_registry)
        elt_types.append(elt_type)

    type_structure = types.TypeStructure(types.StructureType.TUPLE, elt_types)

    return type_structure


def _build_structure(
    type_def: _TypeDefinition, type_registry: _TypeRegistry
) -> Optional[types.TypeStructure]:
    """
    Build a TypeStructure instance from a structure definition within a
    type definition.

    Args:
        type_def: A type definition
        type_registry: A type registry, used to look up type references

    Returns:
        The structure instance, or None if the type definition did not define
        any structure.
    """
    structures = {
        structure_type: type_def[structure_type.name.lower()]
        for structure_type in types.StructureType
        if structure_type.name.lower() in type_def
    }

    if len(structures) > 1:
        raise TooManyTypeStructuresError(list(structures))

    structure_type, structure_def = next(iter(structures.items()), (None, None))

    # Asserts below are for mypy; they reflect data structure (e.g. YAML
    # structure), which is something which would be enforced in JSON-Schema.
    # This code processes and is dependent upon proper data structure.
    if structure_type is types.StructureType.LIST:  # noqa: E721
        assert isinstance(structure_def, (Mapping, str))
        structure = _build_list_structure(structure_def, type_registry)
    elif structure_type is types.StructureType.MAPPING:  # noqa: E721
        assert isinstance(structure_def, (Mapping, Sequence))
        structure = _build_mapping_structure(structure_def, type_registry)
    elif structure_type is types.StructureType.TUPLE:  # noqa: E721
        assert isinstance(structure_def, Iterable)
        structure = _build_tuple_structure(structure_def, type_registry)
    else:
        structure = None

    return structure


def _build_member_types(
    member_type_defs: Iterable[_NestedTypeDefOrRef], type_registry: _TypeRegistry
) -> list[types.Type]:
    """
    Create the member types described by a union type definition.

    Args:
        member_type_defs: The member type definitions of the union
        type_registry: A type registry, used to look up type references

    Returns:
        A list of Type instances which are the member types
    """
    member_types = []

    for member_type_def in member_type_defs:
        member_type = build_or_get_type(member_type_def, type_registry)

        member_types.append(member_type)

    return member_types


def build_or_get_type(
    type_def: _NestedTypeDefOrRef, type_registry: _TypeRegistry
) -> types.Type:
    """
    Useful in nested contexts where either a nested anonymous type definition
    or the name of a registered type is allowed.  This function will return the
    registered type if type_def is just a name (string), otherwise it will
    treat type_def as a spec for an anonymous type, build that, and return it.

    Args:
        type_def: A type name or definition
        type_registry: The type registry as has been built so far

    Returns:
        A Type instance
    """
    if isinstance(type_def, str):
        type_ = type_registry[type_def]

    else:
        type_ = build_type(None, type_def, type_registry)

    return type_


def build_type(
    type_name: Optional[str],
    type_def: Optional[_TypeDefinition],
    type_registry: _TypeRegistry,
) -> types.Type:
    """
    Create a Type instance from a name and a type definition.

    Args:
        type_name: The type name; may be None if creating an anonymous type
        type_def: The type definition; may be None if creating a simple type
        type_registry: A type registry, used to look up type references

    Returns:
        A Type instance
    """
    if type_def is None:
        # Simple type
        member_type_defs = super_type_name = structure = None

    else:
        # Non-simple type
        member_type_defs = type_def.get("union")
        super_type_name = type_def.get("is_a")
        structure = _build_structure(type_def, type_registry)

    if super_type_name:
        super_type = type_registry[super_type_name]
    else:
        super_type = None

    if member_type_defs:
        member_types = _build_member_types(member_type_defs, type_registry)
    else:
        member_types = None

    # Sanity checks
    if super_type and (structure or member_types):
        raise DioptraTypeError("Structure/union types can't have a super-type")

    if structure and member_types:
        raise DioptraTypeError("A type may not be both structured and union")

    type_: types.Type
    if structure:
        type_ = types.StructuredType(structure, type_name)

    elif member_types:
        type_ = types.UnionType(member_types, type_name)

    else:  # else, a simple type
        if super_type and not isinstance(super_type, types.SimpleType):
            raise NonSimpleSuperTypeError(super_type_name)

        # Here, super_type must either be null or an instance of SimpleType
        # (the negation of the above if condition).  I.e. it satisfies
        # Optional[SimpleType].  But mypy can't seem to grok this.  And I
        # don't think there's a simple type guard you can use.  So just cast.
        super_type = cast(Optional[types.SimpleType], super_type)

        if type_name is None:
            raise AnonymousSimpleTypeError()

        type_ = types.SimpleType(type_name, super_type)

    return type_


def get_dependency_types(  # noqa: C901
    type_def: Optional[Union[_TypeDefinition, str]]
) -> Iterator[str]:
    """
    Search the given type definition and generate all references to other
    types found within.

    Args:
        type_def: A type definition

    Yields:
        Dependency type names
    """
    if isinstance(type_def, str):
        yield type_def

    elif type_def is not None:
        # Check structures
        for struct_type in types.StructureType:
            struct_def = type_def.get(struct_type.name.lower())
            if struct_def:
                break
        else:
            struct_def = None

        if struct_def:
            if struct_type is types.StructureType.MAPPING:  # noqa: E721
                if isinstance(struct_def, Mapping):
                    for prop_type in struct_def.values():
                        yield from get_dependency_types(prop_type)

                else:
                    # [key type, value type] list
                    yield from get_dependency_types(struct_def[0])
                    yield from get_dependency_types(struct_def[1])

            elif struct_type is types.StructureType.LIST:  # noqa: E721
                yield from get_dependency_types(struct_def)

            else:
                # Must be a tuple structure
                for elt in struct_def:
                    yield from get_dependency_types(elt)

        # Check unions
        union = type_def.get("union")
        if union:
            for elt in union:
                yield from get_dependency_types(elt)

        # Check supertype
        super_type = type_def.get("is_a")
        if super_type:
            yield super_type


def get_sorted_types(type_defs: Mapping[str, _TypeDefinition]) -> list[str]:
    """
    Find a topological sorted list of type names for the given type
    definitions.

    Args:
        type_defs: Type definitions, as a mapping from type name to type
            definition.

    Returns:
        A list of type names
    """
    topo_sorter: graphlib.TopologicalSorter = graphlib.TopologicalSorter()

    for type_name, type_def in type_defs.items():
        topo_sorter.add(type_name)

        for dep_type in get_dependency_types(type_def):
            if dep_type not in type_defs and dep_type not in BUILTIN_TYPES:
                raise TypeNotFoundError(dep_type, context_type_name=type_name)

            topo_sorter.add(type_name, dep_type)

    try:
        sorted_types = list(topo_sorter.static_order())
    except graphlib.CycleError as e:
        raise TypeReferenceCycleError(e.args[1]) from e

    return sorted_types


def build_type_registry(
    type_defs: Mapping[str, _TypeDefinition]
) -> Mapping[str, types.Type]:
    """
    Create a type registry from a set of type definitions.

    Args:
        type_defs: The type definitions, a mapping type name to type definition

    Returns:
        A type registry, as a mapping from type name to Type instance.  The
        mapping will also include builtin types which are not explicitly
        defined in the given type definitions, so that the registry includes
        all known types.
    """
    # start with our "builtin" types.  Need a cast because the inferred type
    # of BUILTIN_TYPES is dict[str, SimpleType] since I just have simple types
    # in it for now.  So it's not technically wrong... but I intend to do
    # whatever I want with my copy, including putting other kinds of types into
    # it.
    type_registry = cast(MutableMapping[str, types.Type], BUILTIN_TYPES.copy())

    sorted_type_names = get_sorted_types(type_defs)

    for type_name in sorted_type_names:
        # The topo sorter will add all values to the sorted list regardless of
        # whether they actually label a node (e.g. if a value *only* shows up
        # in a list of prerequisites).  So it is possible to have values in the
        # topo sorted list which are not keys in the type_defs mapping.  These
        # correspond to references to built-in types (not errors) or undefined
        # types (errors).

        if type_name in BUILTIN_TYPES:
            if type_name in type_defs:
                raise BuiltinTypeRedefinitionError(type_name)

        else:
            try:
                type_ = build_type(type_name, type_defs[type_name], type_registry)
            except DioptraTypeError as e:
                if not e.context_type_name:
                    e.context_type_name = type_name
                raise

            type_registry[type_name] = type_

    return type_registry



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""Custom SQLAlchemy data types for the REST API database."""
import datetime
import uuid

from sqlalchemy import DateTime as SADateTime
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.types import CHAR, TypeDecorator


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses CHAR(32), storing as stringified hex
    values.

    Note:
        This implementation is adapted from the following section of the SQLAlchemy
        documentation:
        https://docs.sqlalchemy.org/en/20/core/custom_types.html#backend-agnostic-guid-type
    """

    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())

        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value

        elif dialect.name == "postgresql":
            return str(value)

        else:
            if not isinstance(value, uuid.UUID):
                return uuid.UUID(value).hex

            else:
                # hexstring
                return value.hex

    def process_result_value(self, value, dialect):
        if value is None:
            return value

        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)

            return value


class TZDateTime(TypeDecorator):
    """Timezone-aware timestamps stored as timezone-naive UTC timestamps.

    Converts incoming timestamps with timezone information to UTC and then strips out
    the timezone information before storing in the database. This implementation is
    preferred because it is compatible with all database backends.

    Note:
        This implementation is adapted from the following section of the SQLAlchemy
        documentation:
        https://docs.sqlalchemy.org/en/20/core/custom_types.html#store-timezone-aware-timestamps-as-timezone-naive-utc
    """

    impl = SADateTime
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is None:
            return value

        if not value.tzinfo or value.tzinfo.utcoffset(value) is None:
            raise TypeError("tzinfo is required")

        value = value.astimezone(datetime.timezone.utc).replace(tzinfo=None)

        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = value.replace(tzinfo=datetime.timezone.utc)

        return value



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The SQLAlchemy Metadata and Flask integration for the Dioptra database.

For further information about the SQLAlchemy MetaData class, see:
https://docs.sqlalchemy.org/en/20/core/metadata.html

For further information about configuring the SQLAlchemy constraint naming conventions,
see:
https://docs.sqlalchemy.org/en/20/core/constraints.html#constraint-naming-conventions
"""
import datetime
import uuid
from sqlite3 import Connection as SQLite3Connection
from typing import Annotated, Any, Optional

from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import JSON, BigInteger, Boolean, Integer, MetaData, String, Text, event
from sqlalchemy.engine import Engine
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import DeclarativeBase, MappedAsDataclass, mapped_column

from .custom_types import GUID, TZDateTime

intpk = Annotated[
    int, mapped_column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True)
]
bigint = Annotated[int, mapped_column(BigInteger().with_variant(Integer, "sqlite"))]
bigintnovariant = Annotated[int, mapped_column(BigInteger())]
guid = Annotated[uuid.UUID, mapped_column(GUID())]
datetimetz = Annotated[datetime.datetime, mapped_column(TZDateTime())]
optionalbigint = Annotated[
    Optional[int],
    mapped_column(BigInteger().with_variant(Integer, "sqlite"), nullable=True),
]
optionaldatetimetz = Annotated[Optional[datetime.datetime], mapped_column(TZDateTime())]
optionaljson_ = Annotated[Optional[dict[str, Any]], mapped_column(JSON)]
optionalstr = Annotated[Optional[str], mapped_column(Text(), nullable=True)]
optionalstr36 = Annotated[Optional[str], mapped_column(String(36), nullable=True)]
json_ = Annotated[dict[str, Any], mapped_column(MutableDict.as_mutable(JSON))]
str36 = Annotated[str, mapped_column(String(36))]
str255 = Annotated[str, mapped_column(String(255))]
text_ = Annotated[str, mapped_column(Text())]
bool_ = Annotated[bool, mapped_column(Boolean())]

_naming_convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}

metadata_obj = MetaData(naming_convention=_naming_convention)


@event.listens_for(Engine, "connect")
def _set_sqlite_pragma(
    dbapi_connection: SQLite3Connection, connection_record: Any
) -> None:
    """Set 'PRAGMA foreign_keys=ON' for each new SQLite connection."""
    if isinstance(dbapi_connection, SQLite3Connection):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON;")
        cursor.close()


class Base(DeclarativeBase, MappedAsDataclass):
    """The base ORM class."""

    metadata = metadata_obj


db = SQLAlchemy(model_class=Base)



import os
from contextlib import contextmanager
from typing import Iterator


@contextmanager
def env_vars(env_updates: dict[str, str]) -> Iterator[None]:
    """Create a context for temporarily updating environment variables.

    Args:
        env_updates: A dictionary that maps the environment variables (keys) to a new
            value. Both the keys and values must be strings.

    Examples:
        The example below shows how to use env_vars to set/update the value of the
        environment variable __A within a context block. First, import the os module
        and check the current value of __A (we provide a a default value because this
        environment variable should not exist).

        >>> import os
        >>> print(os.getenv("__A", "default value of __A"))
        default value of __A

        Next, set up a with statement using env_vars to set a new value for __A confirm
        that the update worked.

        >>> with env_vars({"__A": "new value of __A"}):
        ...     print(os.getenv("__A", "default value of A"))
        new value of __A

        Finally, confirm that __A has been restored to its default value after exiting
        the with statement.

        >>> print(os.getenv("__A", "default value of __A"))
        default value of __A

    Note:
        This code is adapted from https://stackoverflow.com/a/34333710.
    """
    original_env = os.environ.copy()
    os.environ.update(env_updates)

    try:
        yield

    finally:
        os.environ.clear()
        os.environ.update(original_env)



"""Mark existing database tables as 'legacy'

Revision ID: ef1936dcc5eb
Revises: 7086ad09e884
Create Date: 2024-02-09 10:57:16.378149

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "ef1936dcc5eb"
down_revision = "7086ad09e884"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("experiments", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_experiments_name"))

    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_jobs_status"))
        batch_op.drop_index(batch_op.f("ix_jobs_queue_id"))
        batch_op.drop_index(batch_op.f("ix_jobs_mlflow_run_id"))
        batch_op.drop_index(batch_op.f("ix_jobs_experiment_id"))
        batch_op.drop_constraint(batch_op.f("fk_jobs_status_job_statuses"))
        batch_op.drop_constraint(batch_op.f("fk_jobs_queue_id_queues"))
        batch_op.drop_constraint(batch_op.f("fk_jobs_experiment_id_experiments"))

    with op.batch_alter_table("queues", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_queues_name"))

    with op.batch_alter_table("queue_locks", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("fk_queue_locks_queue_id_queues"))

    op.rename_table("experiments", "legacy_experiments")
    op.rename_table("jobs", "legacy_jobs")
    op.rename_table("queues", "legacy_queues")
    op.rename_table("users", "legacy_users")
    op.rename_table("job_statuses", "legacy_job_statuses")
    op.rename_table("queue_locks", "legacy_queue_locks")

    with op.batch_alter_table("legacy_experiments", schema=None) as batch_op:
        batch_op.create_index("ix_legacy_experiments_name", ["name"], unique=1)

    with op.batch_alter_table("legacy_jobs", schema=None) as batch_op:
        batch_op.create_index("ix_legacy_jobs_status", ["status"], unique=False)
        batch_op.create_index("ix_legacy_jobs_queue_id", ["queue_id"], unique=False)
        batch_op.create_index(
            "ix_legacy_jobs_mlflow_run_id", ["mlflow_run_id"], unique=False
        )
        batch_op.create_index(
            "ix_legacy_jobs_experiment_id", ["experiment_id"], unique=False
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_legacy_jobs_status_legacy_job_statuses"),
            "legacy_job_statuses",
            ["status"],
            ["status"],
        )
        batch_op.create_foreign_key(
            "fk_legacy_jobs_queue_id_legacy_queues",
            "legacy_queues",
            ["queue_id"],
            ["queue_id"],
        )
        batch_op.create_foreign_key(
            "fk_legacy_jobs_experiment_id_legacy_experiments",
            "legacy_experiments",
            ["experiment_id"],
            ["experiment_id"],
        )

    with op.batch_alter_table("legacy_queues", schema=None) as batch_op:
        batch_op.create_index("ix_legacy_queues_name", ["name"], unique=1)

    with op.batch_alter_table("legacy_queue_locks", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_legacy_queue_locks_queue_id_legacy_queues",
            "legacy_queues",
            ["queue_id"],
            ["queue_id"],
        )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("legacy_experiments", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_legacy_experiments_name"))

    with op.batch_alter_table("legacy_jobs", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_legacy_jobs_status"))
        batch_op.drop_index(batch_op.f("ix_legacy_jobs_queue_id"))
        batch_op.drop_index(batch_op.f("ix_legacy_jobs_mlflow_run_id"))
        batch_op.drop_index(batch_op.f("ix_legacy_jobs_experiment_id"))
        batch_op.drop_constraint(
            batch_op.f("fk_legacy_jobs_status_legacy_job_statuses")
        )
        batch_op.drop_constraint(batch_op.f("fk_legacy_jobs_queue_id_legacy_queues"))
        batch_op.drop_constraint(
            batch_op.f("fk_legacy_jobs_experiment_id_legacy_experiments")
        )

    with op.batch_alter_table("legacy_queues", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_legacy_queues_name"))

    with op.batch_alter_table("legacy_queue_locks", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_legacy_queue_locks_queue_id_legacy_queues")
        )

    op.rename_table("legacy_experiments", "experiments")
    op.rename_table("legacy_jobs", "jobs")
    op.rename_table("legacy_queues", "queues")
    op.rename_table("legacy_users", "users")
    op.rename_table("legacy_job_statuses", "job_statuses")
    op.rename_table("legacy_queue_locks", "queue_locks")

    with op.batch_alter_table("experiments", schema=None) as batch_op:
        batch_op.create_index("ix_experiments_name", ["name"], unique=1)

    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.create_index("ix_jobs_status", ["status"], unique=False)
        batch_op.create_index("ix_jobs_queue_id", ["queue_id"], unique=False)
        batch_op.create_index("ix_jobs_mlflow_run_id", ["mlflow_run_id"], unique=False)
        batch_op.create_index("ix_jobs_experiment_id", ["experiment_id"], unique=False)
        batch_op.create_foreign_key(
            batch_op.f("fk_jobs_status_job_statuses"),
            "job_statuses",
            ["status"],
            ["status"],
        )
        batch_op.create_foreign_key(
            "fk_jobs_queue_id_queues", "queues", ["queue_id"], ["queue_id"]
        )
        batch_op.create_foreign_key(
            "fk_jobs_experiment_id_experiments",
            "experiments",
            ["experiment_id"],
            ["experiment_id"],
        )

    with op.batch_alter_table("queues", schema=None) as batch_op:
        batch_op.create_index("ix_queues_name", ["name"], unique=1)

    with op.batch_alter_table("queue_locks", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_queue_locks_queue_id_queues",
            "queues",
            ["queue_id"],
            ["queue_id"],
        )
    # ### end Alembic commands ###



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for Model resources."""
from __future__ import annotations

import uuid
from typing import cast
from urllib.parse import unquote

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import inject
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.restapi.routes import V1_MODELS_ROUTE
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.schemas import IdStatusResponseSchema
from dioptra.restapi.v1.shared.drafts.controller import (
    generate_resource_drafts_endpoint,
    generate_resource_drafts_id_endpoint,
    generate_resource_id_draft_endpoint,
)
from dioptra.restapi.v1.shared.snapshots.controller import (
    generate_resource_snapshots_endpoint,
    generate_resource_snapshots_id_endpoint,
)
from dioptra.restapi.v1.shared.tags.controller import (
    generate_resource_tags_endpoint,
    generate_resource_tags_id_endpoint,
)

from .schema import (
    ModelGetQueryParameters,
    ModelMutableFieldsSchema,
    ModelPageSchema,
    ModelSchema,
    ModelVersionGetQueryParameters,
    ModelVersionMutableFieldsSchema,
    ModelVersionPageSchema,
    ModelVersionSchema,
)
from .service import (
    MODEL_RESOURCE_TYPE,
    MODEL_SEARCHABLE_FIELDS,
    ModelIdService,
    ModelIdVersionsNumberService,
    ModelIdVersionsService,
    ModelService,
)

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace("Models", description="Models endpoint")


@api.route("/")
class ModelEndpoint(Resource):
    @inject
    def __init__(self, model_service: ModelService, *args, **kwargs) -> None:
        """Initialize the model resource.

        All arguments are provided via dependency injection.

        Args:
            model_service: A ModelService object.
        """
        self._model_service = model_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=ModelGetQueryParameters, api=api)
    @responds(schema=ModelPageSchema, api=api)
    def get(self):
        """Gets a list of all Model resources."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Models", request_type="GET"
        )

        parsed_query_params = request.parsed_query_params  # noqa: F841
        group_id = parsed_query_params["group_id"]
        search_string = unquote(parsed_query_params["search"])
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]
        sort_by_string = parsed_query_params["sort_by"]
        descending = parsed_query_params["descending"]

        models, total_num_models = self._model_service.get(
            group_id=group_id,
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
            sort_by_string=sort_by_string,
            descending=descending,
            log=log,
        )
        return utils.build_paging_envelope(
            "models",
            build_fn=utils.build_model,
            data=models,
            group_id=group_id,
            query=search_string,
            draft_type=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_models,
            sort_by=sort_by_string,
            descending=descending,
        )

    @login_required
    @accepts(schema=ModelSchema, api=api)
    @responds(schema=ModelSchema, api=api)
    def post(self):
        """Creates a Model resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Model", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # noqa: F841
        model = self._model_service.create(
            name=parsed_obj["name"],
            description=parsed_obj["description"],
            group_id=parsed_obj["group_id"],
            log=log,
        )
        return utils.build_model(model)


@api.route("/<int:id>")
@api.param("id", "ID for the Model resource.")
class ModelIdEndpoint(Resource):
    @inject
    def __init__(self, model_id_service: ModelIdService, *args, **kwargs) -> None:
        """Initialize the model resource.

        All arguments are provided via dependency injection.

        Args:
            model_id_service: A ModelIdService object.
        """
        self._model_id_service = model_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=ModelSchema, api=api)
    def get(self, id: int):
        """Gets a Model resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Model", request_type="GET", id=id
        )
        model = cast(
            models.MlModel,
            self._model_id_service.get(id, error_if_not_found=True, log=log),
        )
        return utils.build_model(model)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int):
        """Deletes a Model resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Model", request_type="DELETE", id=id
        )
        return self._model_id_service.delete(model_id=id, log=log)

    @login_required
    @accepts(schema=ModelMutableFieldsSchema, api=api)
    @responds(schema=ModelSchema, api=api)
    def put(self, id: int):
        """Modifies a Model resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Model", request_type="PUT", id=id
        )
        parsed_obj = request.parsed_obj  # type: ignore # noqa: F841
        model = cast(
            models.MlModel,
            self._model_id_service.modify(
                id,
                name=parsed_obj["name"],
                description=parsed_obj["description"],
                error_if_not_found=True,
                log=log,
            ),
        )
        return utils.build_model(model)


@api.route("/<int:id>/versions")
@api.param("id", "ID for the Models resource.")
class ModelIdVersionsEndpoint(Resource):
    @inject
    def __init__(
        self, model_id_versions_service: ModelIdVersionsService, *args, **kwargs
    ) -> None:
        """Initialize the model resource.

        All arguments are provided via dependency injection.

        Args:
            model_id_versions_service: A ModelIdVersionsService object.
        """
        self._model_id_versions_service = model_id_versions_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=ModelVersionGetQueryParameters, api=api)
    @responds(schema=ModelVersionPageSchema, api=api)
    def get(self, id: int):
        """Gets a list of versions of this Model resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Model", request_type="GET"
        )

        parsed_query_params = request.parsed_query_params  # type: ignore
        search_string = unquote(parsed_query_params["search"])
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]

        versions, total_num_versions = cast(
            tuple[list[utils.ModelWithVersionDict], int],
            self._model_id_versions_service.get(
                model_id=id,
                search_string=search_string,
                page_index=page_index,
                page_length=page_length,
                error_if_not_found=True,
                log=log,
            ),
        )
        return utils.build_paging_envelope(
            f"models/{id}/versions",
            build_fn=utils.build_model_version,
            data=versions,
            group_id=None,
            query=search_string,
            draft_type=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_versions,
        )

    @login_required
    @accepts(schema=ModelVersionSchema, api=api)
    @responds(schema=ModelVersionSchema, api=api)
    def post(self, id: int):
        """Creates a new version of the model resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Model", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # type: ignore
        model = self._model_id_versions_service.create(
            id,
            description=parsed_obj["description"],
            artifact_id=parsed_obj["artifact_id"],
            log=log,
        )
        return utils.build_model_version(model)


@api.route("/<int:id>/versions/<int:versionNumber>")
@api.param("id", "ID for the Model resource.")
@api.param("versionNumber", "Version number for the Model resource.")
class ModelIdVersionsNumberEndpoint(Resource):
    @inject
    def __init__(
        self,
        model_id_versions_number_service: ModelIdVersionsNumberService,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the model resource.

        All arguments are provided via dependency injection.

        Args:
            model_id_versions_number_service: A ModelIdVersionsNumberService object.
        """
        self._model_id_versions_number_service = model_id_versions_number_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=ModelVersionSchema, api=api)
    def get(self, id: int, versionNumber: int):
        """Gets a specific version of a model by version number."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Models", request_type="GET"
        )
        model = cast(
            utils.ModelWithVersionDict,
            self._model_id_versions_number_service.get(
                id, version_number=versionNumber, error_if_not_found=True, log=log
            ),
        )
        return utils.build_model_version(model)

    @login_required
    @accepts(schema=ModelVersionMutableFieldsSchema, api=api)
    @responds(schema=ModelVersionSchema, api=api)
    def put(self, id: int, versionNumber: int):
        """Modifies specific version of the model resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Model", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # type: ignore
        model = cast(
            utils.ModelWithVersionDict,
            self._model_id_versions_number_service.modify(
                id,
                versionNumber,
                description=parsed_obj["description"],
                error_if_not_found=True,
                log=log,
            ),
        )
        return utils.build_model_version(model)


ModelDraftResource = generate_resource_drafts_endpoint(
    api,
    resource_name=MODEL_RESOURCE_TYPE,
    route_prefix=V1_MODELS_ROUTE,
    request_schema=ModelSchema,
)
ModelDraftIdResource = generate_resource_drafts_id_endpoint(
    api,
    resource_name=MODEL_RESOURCE_TYPE,
    request_schema=ModelSchema(exclude=["groupId"]),
)
ModelIdDraftResource = generate_resource_id_draft_endpoint(
    api,
    resource_name=MODEL_RESOURCE_TYPE,
    request_schema=ModelSchema(exclude=["groupId"]),
)

ModelSnapshotsResource = generate_resource_snapshots_endpoint(
    api=api,
    resource_model=models.MlModel,
    resource_name=MODEL_RESOURCE_TYPE,
    route_prefix=V1_MODELS_ROUTE,
    searchable_fields=MODEL_SEARCHABLE_FIELDS,
    page_schema=ModelPageSchema,
    build_fn=utils.build_model,
)
ModelSnapshotsIdResource = generate_resource_snapshots_id_endpoint(
    api=api,
    resource_model=models.MlModel,
    resource_name=MODEL_RESOURCE_TYPE,
    response_schema=ModelSchema,
    build_fn=utils.build_model,
)

ModelTagsResource = generate_resource_tags_endpoint(
    api=api,
    resource_name=MODEL_RESOURCE_TYPE,
)
ModelTagsIdResource = generate_resource_tags_id_endpoint(
    api=api,
    resource_name=MODEL_RESOURCE_TYPE,
)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .bounding_box_coordinates import BoundingBoxCoordinates, BoundingBoxesBatchedGrid


class NumpyBoundingBoxCoordinates(BoundingBoxCoordinates):
    def __init__(self, grid_shape: tuple[int, int]) -> None:
        self._grid_shape = grid_shape

    @property
    def cell_height(self) -> float:
        return 1 / self._grid_shape[0]

    @property
    def cell_width(self) -> float:
        return 1 / self._grid_shape[1]

    @property
    def cell_ncol(self) -> int:
        return self._grid_shape[1]

    @property
    def cell_nrow(self) -> int:
        return self._grid_shape[0]

    def apply_constraint_one_object_per_cell(
        self,
        bboxes_cell_xywh: npt.NDArray,
        bboxes_cell_ij: npt.NDArray,
        bboxes_labels: npt.NDArray,
    ) -> tuple[npt.NDArray, npt.NDArray, npt.NDArray]:
        pruned_bboxes_cell_ij, pruning_indices = np.unique(
            bboxes_cell_ij, return_index=True, axis=0
        )
        pruned_bboxes_cell_xywh = bboxes_cell_xywh[pruning_indices]
        pruned_bboxes_labels = bboxes_labels[pruning_indices]

        return (
            pruned_bboxes_cell_xywh.astype("float32"),
            pruned_bboxes_cell_ij.astype("int32"),
            pruned_bboxes_labels.astype("int32"),
        )

    def find_no_obj_cell_ij(self, bboxes_cell_ij: npt.NDArray) -> npt.NDArray:
        cell_ij_set: set[tuple[int, int]] = set(
            [(i, j) for i in range(self.cell_nrow) for j in range(self.cell_ncol)]
        )
        cell_ij_seen: set[tuple[int, int]] = set(
            [(int(x[0]), int(x[1])) for x in bboxes_cell_ij.tolist()]
        )
        no_obj_cell_ij = sorted(list(cell_ij_set - cell_ij_seen))

        return np.array(no_obj_cell_ij, dtype="int32")

    def find_bbox_cell_ij(
        self, x_center: npt.NDArray, y_center: npt.NDArray
    ) -> npt.NDArray:
        cell_h: float = self.cell_height
        cell_w: float = self.cell_width
        max_i: int = self.cell_nrow
        max_j: int = self.cell_ncol

        i: npt.NDArray = np.min(
            np.stack(
                [
                    (y_center / cell_h).astype("int32"),
                    np.full(shape=y_center.shape, fill_value=max_i, dtype="int32"),
                ],
                axis=-1,
            ),
            axis=-1,
        )
        j: npt.NDArray = np.min(
            np.stack(
                [
                    (x_center / cell_w).astype("int32"),
                    np.full(shape=x_center.shape, fill_value=max_j, dtype="int32"),
                ],
                axis=-1,
            ),
            axis=-1,
        )

        bbox_cell_ij: npt.NDArray = np.stack([i, j], axis=-1).astype("int32")

        return bbox_cell_ij

    def find_bbox_cell_xy(
        self, x_center: npt.NDArray, y_center: npt.NDArray, bboxes_cell_ij: npt.NDArray
    ) -> tuple[npt.NDArray, npt.NDArray]:
        cell_h: float = self.cell_height
        cell_w: float = self.cell_width

        cell_x: npt.NDArray = np.min(
            np.stack(
                [
                    (x_center / cell_w).astype("float32")
                    - bboxes_cell_ij[..., 1].astype("float32"),
                    np.full(shape=x_center.shape, fill_value=1.0, dtype="float32"),
                ],
                axis=-1,
            ),
            axis=-1,
        )
        cell_y: npt.NDArray = np.min(
            np.stack(
                [
                    (y_center / cell_h).astype("float32")
                    - bboxes_cell_ij[..., 0].astype("float32"),
                    np.full(shape=y_center.shape, fill_value=1.0, dtype="float32"),
                ],
                axis=-1,
            ),
            axis=-1,
        )

        return cell_x, cell_y

    def from_corner_to_image_xywh(self, bboxes_corner: npt.NDArray) -> npt.NDArray:
        xmin = bboxes_corner[..., 0]
        ymin = bboxes_corner[..., 1]
        xmax = bboxes_corner[..., 2]
        ymax = bboxes_corner[..., 3]

        x = (xmin + xmax) / 2.0
        y = (ymin + ymax) / 2.0
        w = xmax - xmin
        h = ymax - ymin

        return np.stack([x, y, w, h], axis=-1)

    def from_image_to_cell_xywh(
        self, bboxes_image_xywh: npt.NDArray
    ) -> tuple[npt.NDArray, npt.NDArray]:
        x_center = bboxes_image_xywh[..., 0]
        y_center = bboxes_image_xywh[..., 1]

        bboxes_cell_ij = self.find_bbox_cell_ij(x_center=x_center, y_center=y_center)
        cell_x, cell_y = self.find_bbox_cell_xy(
            x_center=x_center, y_center=y_center, bboxes_cell_ij=bboxes_cell_ij
        )
        bboxes_cell_xywh = np.stack(
            [cell_x, cell_y, bboxes_image_xywh[..., 2], bboxes_image_xywh[..., 3]],
            axis=-1,
        )

        return bboxes_cell_xywh.astype("float32"), bboxes_cell_ij.astype("int32")

    def from_corner_to_cell_xywh(
        self, bboxes_corner: npt.NDArray
    ) -> tuple[npt.NDArray, npt.NDArray]:
        bboxes_image_xywh: npt.NDArray = self.from_corner_to_image_xywh(
            bboxes_corner=bboxes_corner
        )
        bboxes_cell_xywh, bboxes_cell_ij = self.from_image_to_cell_xywh(
            bboxes_image_xywh=bboxes_image_xywh
        )

        return bboxes_cell_xywh, bboxes_cell_ij

    def from_cell_xywh_to_image_xywh(
        self, bboxes_cell_xywh: npt.NDArray, bboxes_cell_ij: npt.NDArray
    ) -> npt.NDArray:
        num_cells_wh = np.array([self.cell_ncol, self.cell_nrow], dtype="float32")
        bboxes_image_xywh = bboxes_cell_xywh.copy()
        bboxes_image_xywh[..., :2] = (
            bboxes_cell_xywh[..., :2] + bboxes_cell_ij[:, [1, 0]].astype("float32")
        ) / num_cells_wh

        return bboxes_image_xywh

    def from_image_xywh_to_corner(self, bboxes_image_xywh: npt.NDArray) -> npt.NDArray:
        x = bboxes_image_xywh[..., 0]
        y = bboxes_image_xywh[..., 1]
        w = bboxes_image_xywh[..., 2]
        h = bboxes_image_xywh[..., 3]

        x1 = x - (w / 2)
        x2 = x + (w / 2)
        y1 = y - (h / 2)
        y2 = y + (h / 2)

        return np.stack([x1, y1, x2, y2], axis=-1)

    def from_cell_xywh_to_corner(
        self, bboxes_cell_xywh: npt.NDArray, bboxes_cell_ij: npt.NDArray
    ) -> npt.NDArray:
        bboxes_image_xywh = self.from_cell_xywh_to_image_xywh(
            bboxes_cell_xywh=bboxes_cell_xywh, bboxes_cell_ij=bboxes_cell_ij
        )
        bboxes_corner = self.from_image_xywh_to_corner(
            bboxes_image_xywh=bboxes_image_xywh
        )

        return bboxes_corner


class NumpyBoundingBoxesBatchedGrid(BoundingBoxesBatchedGrid):
    def __init__(
        self,
        bounding_box_coordinates: NumpyBoundingBoxCoordinates,
    ) -> None:
        self._bbox_coord = bounding_box_coordinates
        self._wh_grid_indices = self._generate_wh_grid_indices(
            grid_shape=(
                bounding_box_coordinates.cell_nrow,
                bounding_box_coordinates.cell_ncol,
            )
        )
        self._num_wh_grid_indices = self._generate_num_wh_grid_indices(
            grid_shape=(
                bounding_box_coordinates.cell_nrow,
                bounding_box_coordinates.cell_ncol,
            )
        )

    @classmethod
    def on_grid_shape(
        cls, grid_shape: tuple[int, int]
    ) -> NumpyBoundingBoxesBatchedGrid:
        bbox_coord: NumpyBoundingBoxCoordinates = NumpyBoundingBoxCoordinates(
            grid_shape=grid_shape
        )

        return cls(bounding_box_coordinates=bbox_coord)

    @property
    def cell_height(self) -> float:
        return self._bbox_coord.cell_height

    @property
    def cell_width(self) -> float:
        return self._bbox_coord.cell_width

    @property
    def cell_ncol(self) -> int:
        return self._bbox_coord.cell_ncol

    @property
    def cell_nrow(self) -> int:
        return self._bbox_coord.cell_nrow

    def embed(
        self, bboxes_corner: npt.NDArray, bboxes_labels: npt.NDArray, n_classes: int
    ) -> tuple[npt.NDArray, npt.NDArray, npt.NDArray, npt.NDArray]:
        bboxes_cell_xywh_grid: npt.NDArray = np.zeros(
            shape=(self.cell_nrow, self.cell_ncol, 4),
            dtype="float32",
        )
        bboxes_labels_grid: npt.NDArray = np.zeros(
            shape=(self.cell_nrow, self.cell_ncol, n_classes),
            dtype="float32",
        )
        bboxes_object_mask: npt.NDArray = np.zeros(
            shape=(self.cell_nrow, self.cell_ncol),
            dtype="float32",
        )
        bboxes_no_object_mask: npt.NDArray = np.ones(
            shape=(self.cell_nrow, self.cell_ncol),
            dtype="float32",
        )

        bboxes_cell_xywh, bboxes_cell_ij = self._bbox_coord.from_corner_to_cell_xywh(
            bboxes_corner=bboxes_corner
        )
        xywh, ij, labels = self._bbox_coord.apply_constraint_one_object_per_cell(
            bboxes_cell_xywh=bboxes_cell_xywh,
            bboxes_cell_ij=bboxes_cell_ij,
            bboxes_labels=bboxes_labels,
        )

        grid_i = np.transpose(ij)[0]
        grid_j = np.transpose(ij)[1]

        bboxes_cell_xywh_grid[grid_i, grid_j] = xywh
        bboxes_labels_grid[grid_i, grid_j, labels] = 1.0
        bboxes_object_mask[grid_i, grid_j] = 1.0
        bboxes_no_object_mask[grid_i, grid_j] = 0.0

        bboxes_cell_xywh_grid = np.expand_dims(bboxes_cell_xywh_grid, axis=-2)

        return (
            bboxes_cell_xywh_grid,
            bboxes_labels_grid,
            bboxes_object_mask,
            bboxes_no_object_mask,
        )

    def from_corner_to_image_xywh(self, bboxes_corner: npt.NDArray) -> npt.NDArray:
        return self._bbox_coord.from_corner_to_image_xywh(bboxes_corner=bboxes_corner)

    def from_image_to_cell_xywh(self, bboxes_image_xywh: npt.NDArray) -> npt.NDArray:
        bboxes_cell_xywh, _ = self._bbox_coord.from_image_to_cell_xywh(
            bboxes_image_xywh=bboxes_image_xywh
        )

        return bboxes_cell_xywh

    def from_corner_to_cell_xywh(self, bboxes_corner: npt.NDArray) -> npt.NDArray:
        bboxes_image_xywh: npt.NDArray = self.from_corner_to_image_xywh(
            bboxes_corner=bboxes_corner
        )
        bboxes_cell_xywh: npt.NDArray = self.from_image_to_cell_xywh(
            bboxes_image_xywh=bboxes_image_xywh
        )

        return bboxes_cell_xywh

    def from_cell_xywh_to_image_xywh(
        self, bboxes_cell_xywh: npt.NDArray, n_bounding_boxes: int
    ) -> npt.NDArray:
        batch_size: int = int(np.shape(bboxes_cell_xywh)[0])

        wh_grid = np.tile(
            self._wh_grid_indices.astype("float32"),
            reps=[batch_size, 1, 1, n_bounding_boxes, 1],
        )
        num_wh_grid = np.tile(
            self._num_wh_grid_indices.astype("float32"),
            reps=[batch_size, 1, 1, n_bounding_boxes, 1],
        )

        bboxes_image_xywh = bboxes_cell_xywh.copy()
        bboxes_image_xywh[..., :2] = (bboxes_cell_xywh[..., :2] + wh_grid) / num_wh_grid

        return bboxes_image_xywh

    def from_image_xywh_to_corner(self, bboxes_image_xywh: npt.NDArray) -> npt.NDArray:
        return self._bbox_coord.from_image_xywh_to_corner(
            bboxes_image_xywh=bboxes_image_xywh
        )

    def from_cell_xywh_to_corner(
        self, bboxes_cell_xywh: npt.NDArray, n_bounding_boxes: int
    ) -> npt.NDArray:
        bboxes_image_xywh = self.from_cell_xywh_to_image_xywh(
            bboxes_cell_xywh=bboxes_cell_xywh,
            n_bounding_boxes=n_bounding_boxes,
        )
        bboxes_corner = self.from_image_xywh_to_corner(
            bboxes_image_xywh=bboxes_image_xywh
        )

        return bboxes_corner

    @staticmethod
    def _generate_wh_grid_indices(grid_shape: tuple[int, int]) -> npt.NDArray:
        w_grid_indices = np.tile(
            np.expand_dims(np.arange(grid_shape[1], dtype="int32"), axis=0),
            reps=[grid_shape[0], 1],
        )
        w_grid_indices = np.reshape(
            w_grid_indices, (1, grid_shape[0], grid_shape[1], 1, 1)
        )

        h_grid_indices = np.transpose(
            np.tile(
                np.expand_dims(np.arange(grid_shape[0], dtype="int32"), axis=0),
                reps=[grid_shape[1], 1],
            )
        )
        h_grid_indices = np.reshape(
            h_grid_indices, (1, grid_shape[0], grid_shape[1], 1, 1)
        )

        wh_grid_indices: npt.NDArray = np.concatenate(
            [w_grid_indices, h_grid_indices],
            axis=-1,
        ).astype("float32")

        return wh_grid_indices

    @staticmethod
    def _generate_num_wh_grid_indices(grid_shape: tuple[int, int]) -> npt.NDArray:
        num_w_grid_indices = np.full(
            shape=(1, grid_shape[0], grid_shape[1], 1, 1),
            fill_value=grid_shape[1],
            dtype="int32",
        )
        num_h_grid_indices = np.full(
            shape=(1, grid_shape[0], grid_shape[1], 1, 1),
            fill_value=grid_shape[0],
            dtype="int32",
        )

        num_wh_grid_indices: npt.NDArray = np.concatenate(
            [num_w_grid_indices, num_h_grid_indices],
            axis=-1,
        ).astype("float32")

        return num_wh_grid_indices



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""A module for the Flask configuration environments.

There are three configuration environments available,

- **dev:** The configuration for interactive development sessions.
- **prod:** The configuration for production environments.
- **test:** The configuration for running unit tests.
"""
from __future__ import annotations

import os
from typing import List, Type


def _set_session_protection(default: str = "strong") -> str:
    """Set the session protection level.

    The session protection level is set by the environment variable
    DIOPTRA_SESSION_PROTECTION. If the environment variable is not set, the default
    value is used.

    Args:
        default: The default session protection level. Must be one of "none", "basic",
            or "strong". Defaults to "strong".

    Returns:
        The session protection level.
    """
    allowed = {"none", "basic", "strong"}
    value = os.getenv("DIOPTRA_SESSION_PROTECTION", default).lower()

    if value not in allowed:
        raise ValueError(
            f"Invalid DIOPTRA_SESSION_PROTECTION value: {value}. "
            f"Allowed values are {allowed}."
        )

    return value


class BaseConfig(object):
    CONFIG_NAME = "base"
    USE_MOCK_EQUIVALENCY = False
    DEBUG = False
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    REMEMBER_COOKIE_NAME = "dioptra_remember_token"
    REMEMBER_COOKIE_DURATION = int(
        os.getenv("DIOPTRA_REMEMBER_COOKIE_DURATION", f"{14 * 86400}")
    )
    REMEMBER_COOKIE_SECURE = os.getenv("DIOPTRA_REMEMBER_COOKIE_SECURE") is not None
    LOGIN_DISABLED = os.getenv("DIOPTRA_LOGIN_DISABLED") is not None
    DIOPTRA_CORS_ORIGIN = os.getenv("DIOPTRA_CORS_ORIGIN", "http://localhost:5173")
    DIOPTRA_PLUGINS_BUCKET = os.getenv("DIOPTRA_PLUGINS_BUCKET", "plugins")
    DIOPTRA_SWAGGER_PATH = os.getenv("DIOPTRA_SWAGGER_PATH", "/")
    DIOPTRA_BASE_URL = os.getenv("DIOPTRA_BASE_URL")


class DevelopmentConfig(BaseConfig):
    CONFIG_NAME = "dev"
    SECRET_KEY = os.getenv("DIOPTRA_DEPLOY_SECRET_KEY", "deploy123")
    DEBUG = True
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    TESTING = False
    SESSION_PROTECTION = _set_session_protection()
    SQLALCHEMY_DATABASE_URI = os.getenv(
        "DIOPTRA_RESTAPI_DEV_DATABASE_URI",
        f"sqlite:///{os.path.join(os.getcwd(), 'dioptra-dev.db')}",
    )


class TestingConfig(BaseConfig):
    CONFIG_NAME = "test"
    SECRET_KEY = os.getenv("DIOPTRA_DEPLOY_SECRET_KEY", "deploy123")
    DEBUG = True
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    TESTING = True
    SESSION_PROTECTION: str | None = None
    SQLALCHEMY_DATABASE_URI = "sqlite://"


class ProductionConfig(BaseConfig):
    CONFIG_NAME = "prod"
    SECRET_KEY = os.getenv("DIOPTRA_DEPLOY_SECRET_KEY", "deploy123")
    DEBUG = False
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    TESTING = False
    SESSION_PROTECTION = _set_session_protection()
    SQLALCHEMY_DATABASE_URI = os.getenv(
        "DIOPTRA_RESTAPI_DATABASE_URI",
        f"sqlite:///{os.path.join(os.getcwd(), 'dioptra.db')}",
    )


class TestingLoginDisabledConfig(TestingConfig):
    CONFIG_NAME = "test_no_login"
    LOGIN_DISABLED = True


class TestingV1Config(TestingConfig):
    CONFIG_NAME = "test_v1"


EXPORT_CONFIGS: List[Type[BaseConfig]] = [
    DevelopmentConfig,
    TestingConfig,
    ProductionConfig,
    TestingV1Config,
    TestingLoginDisabledConfig,
]
config_by_name = {cfg.CONFIG_NAME: cfg for cfg in EXPORT_CONFIGS}



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import contextlib
import functools
from typing import Any, Callable, Iterator, Mapping

import pytest

import dioptra.pyplugs
import dioptra.pyplugs._plugins
import dioptra.task_engine.task_engine
from dioptra.sdk.exceptions.task_engine import (
    IllegalOutputReferenceError,
    IllegalPluginNameError,
    MissingGlobalParametersError,
    MissingTaskPluginNameError,
    NonIterableTaskOutputError,
    OutputNotFoundError,
    StepNotFoundError,
    StepReferenceCycleError,
    UnresolvableReferenceError,
)

_output = None


def capture_return(f: Callable[..., Any]) -> Callable[..., Any]:
    """
    Hacky way to capture a function's return value so unit tests can test it.
    Task graphs have no outputs, so we have no easy way to see what the task
    plugins are returning.

    This function is a decorator which captures the output of the wrapped
    function into the global _output variable.  Of course if more than one
    plugin is invoked, at the end, _output will contain the return value for
    whichever plugin was invoked last.
    """

    @functools.wraps(f)
    def wrapper(*args: Any, **kwargs: Mapping[str, Any]) -> Any:
        global _output
        _output = f(*args, **kwargs)
        return _output

    return wrapper


@capture_return
def add(a: Any, b: Any) -> Any:
    """Simple function to register with pyplugs, for testing"""
    return a + b


@capture_return
def square(n: Any) -> Any:
    """Simple function to register with pyplugs, for testing"""
    return n * n


@capture_return
def addsub(a: Any, b: Any) -> Any:
    """
    Simple function returning multiple values, to register with pyplugs, for
    testing
    """
    return a + b, a - b


@capture_return
def hello() -> str:
    """
    Simple function which takes no args, to register with pyplugs, for testing
    """
    return "hello"


@contextlib.contextmanager
def pyplugs_register(*funcs: Callable[..., Any]) -> Iterator[None]:
    """
    A contextmanager which registers the given function(s) with pyplugs, and
    then un-registers them again to clean up after itself.

    Args:
        *funcs: An iterable of functions (which must support iteration over
            the values multiple times).
    """
    try:
        for func in funcs:
            dioptra.pyplugs.register(func)

        yield

    finally:
        # I don't know of any other way to unregister a plugin, so this uses
        # non-public API.
        plugin_registry = dioptra.pyplugs._plugins._PLUGINS

        # Just search all the plugins; skip splitting up func.__module__ as
        # pyplugs does.  I think there will be relatively few functions
        # registered, so an exhaustive search like this will be plenty fast
        # enough and perhaps simpler to write?  If any of the sub-dicts empties
        # out, the entry in the parent dict which refers to it, is also
        # removed.  It makes for a cleaner clean!
        packages_to_del = []
        for package_name, package in plugin_registry.items():
            plugins_to_del = []
            for plugin_name, plugin in package.items():
                funcs_to_del = []
                for func_name, plugin_info in plugin.items():
                    if plugin_info.func in funcs:
                        funcs_to_del.append(func_name)

                for func_name in funcs_to_del:
                    del plugin[func_name]

                if not plugin:
                    plugins_to_del.append(plugin_name)

            for plugin_name in plugins_to_del:
                del package[plugin_name]

            if not package:
                packages_to_del.append(package_name)

        for package_name in packages_to_del:
            del plugin_registry[package_name]


def require_plugins(
    *funcs: Callable[..., Any]
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Decorator which causes the given plugins to be registered before the
    function runs, and unregistered after it returns.

    Args:
        *funcs: The plugin functions to register

    Returns:
        A wrap function
    """

    def wrap(f):
        def wrapper(*args, **kwargs):
            with pyplugs_register(*funcs):
                return f(*args, **kwargs)

        return wrapper

    return wrap


@require_plugins(add)
def test_single_call_positional() -> None:
    desc = {
        "tasks": {"add": {"plugin": "tests.unit.task_engine.test_task_engine.add"}},
        "graph": {"add": {"add": [1, 1]}},
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert _output == 2


@require_plugins(square)
def test_single_call_positional_nolist() -> None:
    desc = {
        "tasks": {
            "square": {"plugin": "tests.unit.task_engine.test_task_engine.square"}
        },
        "graph": {"step1": {"square": 2}},
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert _output == 4


@require_plugins(add)
def test_single_call_keyword() -> None:
    desc = {
        "tasks": {"add": {"plugin": "tests.unit.task_engine.test_task_engine.add"}},
        "graph": {"step1": {"add": {"a": 1, "b": 1}}},
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert _output == 2


@require_plugins(add)
def test_single_call_mixed_positional_keyword() -> None:
    desc = {
        "tasks": {"add": {"plugin": "tests.unit.task_engine.test_task_engine.add"}},
        "graph": {"step1": {"task": "add", "args": 1, "kwargs": {"b": 1}}},
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert _output == 2


@require_plugins(hello)
def test_single_call_no_args_positional() -> None:
    desc = {
        "tasks": {"hello": {"plugin": "tests.unit.task_engine.test_task_engine.hello"}},
        "graph": {"step1": {"hello": []}},
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert _output == "hello"


@require_plugins(hello)
def test_single_call_no_args_keyword() -> None:
    desc = {
        "tasks": {"hello": {"plugin": "tests.unit.task_engine.test_task_engine.hello"}},
        "graph": {"step1": {"hello": {}}},
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert _output == "hello"


@require_plugins(hello)
def test_single_call_no_args_mixed() -> None:
    desc = {
        "tasks": {"hello": {"plugin": "tests.unit.task_engine.test_task_engine.hello"}},
        "graph": {"step1": {"task": "hello"}},
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert _output == "hello"


@require_plugins(add)
def test_globals_dict_nodefault() -> None:
    desc = {
        "parameters": {"global_in": {"type": "integer"}},
        "tasks": {"add": {"plugin": "tests.unit.task_engine.test_task_engine.add"}},
        "graph": {"step1": {"add": [1, "$global_in"]}},
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {"global_in": 2})

    assert _output == 3

    with pytest.raises(MissingGlobalParametersError) as e:
        dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert "global_in" in e.value.parameter_names


@require_plugins(add)
def test_globals_dict_default() -> None:
    desc = {
        "parameters": {"global_in": 1},
        "tasks": {"add": {"plugin": "tests.unit.task_engine.test_task_engine.add"}},
        "graph": {"step1": {"add": [1, "$global_in"]}},
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {"global_in": 2})

    assert _output == 3

    dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert _output == 2


@require_plugins(addsub)
def test_task_nonlist_output() -> None:
    desc = {
        "tasks": {
            "addsub": {
                "plugin": "tests.unit.task_engine.test_task_engine.addsub",
                "outputs": {"value": "sometype"},
            }
        },
        "graph": {"step1": {"addsub": [1, 2]}},
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert _output == (3, -1)


@require_plugins(add, addsub)
def test_task_list_output() -> None:
    desc = {
        "tasks": {
            "addsub": {
                "plugin": "tests.unit.task_engine.test_task_engine.addsub",
                "outputs": [{"sum": "sometype"}, {"diff": "sometype"}],
            },
            "add": {
                "plugin": "tests.unit.task_engine.test_task_engine.add",
                "outputs": {"value": "sometype"},
            },
        },
        "graph": {
            "step1": {"addsub": [1, 2]},
            # Hard to check how the multi-outputs from "addsub" are actually
            # stored without peeking at the task engine's internal bookkeeping,
            # but I can use those outputs in another step... maybe that will
            # work just as well?
            "step2": {"add": ["$step1.sum", "$step1.diff"]},
        },
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {})

    # (a+b) + (a-b) = 2a
    assert _output == 2


@require_plugins(add, addsub, square)
def test_task_dependencies_ambig() -> None:
    desc = {
        "tasks": {
            "add": {
                "plugin": "tests.unit.task_engine.test_task_engine.add",
                "outputs": {"value": "sometype"},
            },
            "square": {
                "plugin": "tests.unit.task_engine.test_task_engine.square",
                "outputs": {"value": "sometype"},
            },
            # Bad name for a plugin.
            "dependencies": {
                "plugin": "tests.unit.task_engine.test_task_engine.addsub"
            },
        },
        "graph": {
            "step1": {
                # Ensure "dependencies" below isn't mistaken for the same-named
                # plugin defined above!
                "dependencies": ["step2"],
                "add": [3, 4],
            },
            "step2": {"square": 6},
        },
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert _output == 7


@require_plugins(add, addsub, square)
def test_task_task_ambig() -> None:
    desc = {
        "tasks": {
            "add": {
                "plugin": "tests.unit.task_engine.test_task_engine.add",
                "outputs": {"value": "sometype"},
            },
            "square": {
                "plugin": "tests.unit.task_engine.test_task_engine.square",
                "outputs": {"value": "sometype"},
            },
            # Bad name for a plugin.
            "task": {"plugin": "tests.unit.task_engine.test_task_engine.addsub"},
        },
        "graph": {
            "step1": {
                "dependencies": ["step2"],
                # Ensure "task" below isn't mistaken for the same-named
                # plugin defined above!
                "task": "add",
                "args": [3, 4],
            },
            "step2": {"square": 6},
        },
    }

    dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert _output == 7


@require_plugins(add, square)
def test_step_missing_plugin_name() -> None:
    desc = {
        "tasks": {
            "add": {"plugin": "tests.unit.task_engine.test_task_engine.add"},
            "square": {"plugin": "tests.unit.task_engine.test_task_engine.square"},
        },
        "graph": {
            "step1": {
                # Missing the plugin name here
                "dependencies": ["step2"]
            },
            "step2": {"square": 6},
        },
    }

    with pytest.raises(MissingTaskPluginNameError):
        dioptra.task_engine.task_engine.run_experiment(desc, {})


@require_plugins(add, square)
def test_step_cycle() -> None:
    desc = {
        "tasks": {
            "add": {"plugin": "tests.unit.task_engine.test_task_engine.add"},
            "square": {"plugin": "tests.unit.task_engine.test_task_engine.square"},
        },
        "graph": {
            "step1": {"add": [1, 2, "$step2.value"]},
            "step2": {"square": "$step1.value"},
        },
    }

    with pytest.raises(StepReferenceCycleError) as e:
        dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert "step1" in e.value.cycle
    assert "step2" in e.value.cycle


def test_step_not_found_reference() -> None:
    desc = {
        "tasks": {"add": {"plugin": "tests.unit.task_engine.test_task_engine.add"}},
        "graph": {"step1": {"add": [1, "$foo"]}},
    }

    with pytest.raises(UnresolvableReferenceError) as e:
        dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert e.value.reference_name == "foo"


@require_plugins(add)
def test_step_not_found_dependency() -> None:
    desc = {
        "tasks": {"add": {"plugin": "tests.unit.task_engine.test_task_engine.add"}},
        "graph": {"step1": {"add": [1, 2], "dependencies": ["foo"]}},
    }

    with pytest.raises(StepNotFoundError) as e:
        dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert e.value.step_name == "foo"


@require_plugins(add, square)
def test_output_not_found_explicit() -> None:
    desc = {
        "tasks": {
            "add": {
                "plugin": "tests.unit.task_engine.test_task_engine.add",
                "outputs": {"value": "sometype"},
            },
            "square": {"plugin": "tests.unit.task_engine.test_task_engine.square"},
        },
        "graph": {"step1": {"add": [1, 2]}, "step2": {"square": "$step1.foo"}},
    }

    with pytest.raises(OutputNotFoundError) as e:
        dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert e.value.step_name == "step1"
    assert e.value.output_name == "foo"


@require_plugins(add, square)
def test_output_not_found_implicit() -> None:
    desc = {
        "tasks": {
            "add": {"plugin": "tests.unit.task_engine.test_task_engine.add"},
            "square": {"plugin": "tests.unit.task_engine.test_task_engine.square"},
        },
        "graph": {
            "step1": {"add": [1, 2]},
            "step2": {
                # error because plugin "add" did not define any outputs
                "square": "$step1"
            },
        },
    }

    with pytest.raises(UnresolvableReferenceError) as e:
        dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert e.value.reference_name == "step1"


@require_plugins(addsub, square)
def test_output_reference_ambig() -> None:
    desc = {
        "tasks": {
            "addsub": {
                "plugin": "tests.unit.task_engine.test_task_engine.addsub",
                "outputs": ["sum", "diff"],
            },
            "square": {"plugin": "tests.unit.task_engine.test_task_engine.square"},
        },
        "graph": {
            "step1": {"addsub": [1, 2]},
            "step2": {
                # error because plugin "addsub" defines two outputs, so this
                # reference is ambiguous.  It needs to pick one of them.
                "square": "$step1"
            },
        },
    }

    with pytest.raises(IllegalOutputReferenceError) as e:
        dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert e.value.context_step_name == "step2"
    assert e.value.step_name == "step1"


@require_plugins(add)
def test_output_not_iterable() -> None:
    desc = {
        "tasks": {
            "add": {
                "plugin": "tests.unit.task_engine.test_task_engine.add",
                # error: integer addition produces an integer, which is not
                # iterable.
                "outputs": ["value"],
            }
        },
        "graph": {"step1": {"add": [1, 2]}},
    }

    with pytest.raises(NonIterableTaskOutputError) as e:
        dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert e.value.value == 3


def test_illegal_plugin_name() -> None:
    desc = {"tasks": {"add": {"plugin": "foo"}}, "graph": {"step1": {"add": [1, 2]}}}

    with pytest.raises(IllegalPluginNameError) as e:
        dioptra.task_engine.task_engine.run_experiment(desc, {})

    assert e.value.plugin_name == "foo"
